"""Unit tests for gal.py"""
import tempfile

import pytest

from .... import examples as pysal_examples
from ...fileio import FileIO
from ..gal import GalIO


class TesttestGalIO:
    def setup_method(self):
        self.test_file = test_file = pysal_examples.get_path("sids2.gal")
        self.obj = GalIO(test_file, "r")

    def test___init__(self):
        assert self.obj._typ == str

    def test_close(self):
        f = self.obj
        f.close()
        pytest.raises(ValueError, f.read)

    def test_read(self):
        # reading a GAL returns a W
        w = self.obj.read()
        assert w.n == 100
        assert w.sd == pytest.approx(1.5151237573214935)
        assert w.s0 == 462.0
        assert w.s1 == 924.0

    def test_seek(self):
        self.test_read()
        pytest.raises(StopIteration, self.obj.read)
        self.obj.seek(0)
        self.test_read()

    def test_write(self):
        w = self.obj.read()
        f = tempfile.NamedTemporaryFile(suffix=".gal")
        fname = f.name
        f.close()
        o = FileIO(fname, "w")
        o.write(w)
        o.close()
        wnew = FileIO(fname, "r").read()
        assert wnew.pct_nonzero == w.pct_nonzero



# ruff: noqa: F822

import functools as _f

__all__ = [
    "area",
    "bbox",
    "bounding_box",
    "centroid",
    "holes",
    "len",
    "parts",
    "perimeter",
    "segments",
    "vertices",
]


def get_attr(df, geom_col="geometry", inplace=False, attr=None):
    outval = df[geom_col].apply(lambda x: x.__getattribute__(attr))
    if inplace:
        outcol = f"shape_{func.__name__}"  # noqa: F821
        df[outcol] = outval
        return None
    return outval


_doc_template = """
Tabular accessor to grab a geometric object's {n} attribute.

Parameters
----------
df : pandas.DataFrame
    A pandas.Dataframe with a geometry column.
geom_col : str
    The name of the column in ``df`` containing the geometry.
inplace : bool
    A boolean denoting whether to operate on ``df`` inplace or to return a
    pandas.Series contaning the results of the computation. If operating
    inplace, the derived column will be under 'shape_{n}'.

Returns
-------
``None`` if inplace is set to ``True`` and operation is conducted
on ``df`` in memory. Otherwise, returns a pandas.Series.

See Also
--------
For further documentation about the attributes of the object in question, refer
to shape classes in ``pysal.cg.shapes``.
"""

_accessors = {}
for k in __all__:
    _accessors[k] = _f.partial(get_attr, attr=k)
    _accessors[k].__doc__ = _doc_template.format(n=k)

globals().update(_accessors)



# ruff: noqa: N801, N802, N806, N999, SIM115


import csv

from .. import tables

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["csvWrapper"]


class csvWrapper(tables.DataTable):
    """Read a ``.csv`` file in a `DataTable` object.

    Examples
    --------

    >>> import libpysal
    >>> stl = libpysal.examples.load_example('stl')
    >>> file_name = stl.get_path('stl_hom.csv')
    >>> f = libpysal.io.open(file_name,'r')
    >>> y = f.read()
    >>> f.header
    ['WKT',
     'NAME',
     'STATE_NAME',
     'STATE_FIPS',
     'CNTY_FIPS',
     'FIPS',
     'FIPSNO',
     'HR7984',
     'HR8488',
     'HR8893',
     'HC7984',
     'HC8488',
     'HC8893',
     'PO7984',
     'PO8488',
     'PO8893',
     'PE77',
     'PE82',
     'PE87',
     'RDAC80',
     'RDAC85',
     'RDAC90']

    >>> f._spec
    [str,
     str,
     str,
     int,
     int,
     int,
     int,
     float,
     float,
     float,
     int,
     int,
     int,
     int,
     int,
     int,
     float,
     float,
     float,
     float,
     float,
     float]

    """

    __doc__ = tables.DataTable.__doc__  # noqa: A003
    FORMATS = ["csv"]
    READ_MODES = ["r", "Ur", "rU", "U"]
    MODES = READ_MODES[:]

    def __init__(self, *args, **kwargs):
        tables.DataTable.__init__(self, *args, **kwargs)
        self.__idx = {}
        self.__len = None
        self._open()

    def __len__(self):
        return self.__len

    def _open(self):
        self.fileObj = open(self.dataPath, self.mode)
        if self.mode in self.READ_MODES:
            self.dataObj = csv.reader(self.fileObj)
            data = list(self.dataObj)
            if self._determineHeader(data):
                self.header = data.pop(0)
            else:
                self.header = ["field_%d" % i for i in range(len(data[0]))]
            self._spec = self._determineSpec(data)
            self.data = data
            self.fileObj.close()
            self.__len = len(data)

    def _determineHeader(self, data: list) -> bool:
        HEADER = True

        headSpec = self._determineSpec([data[0]])
        restSpec = self._determineSpec(data[1:])

        if headSpec == restSpec:
            HEADER = False

        return HEADER

    @staticmethod
    def _determineSpec(data: list) -> list:
        cols = len(data[0])
        spec = []

        for j in range(cols):
            isInt = True
            isFloat = True

            for row in data:
                val = row[j]
                if not val.strip().replace("-", "").replace(".", "").isdigit():
                    isInt = False
                    isFloat = False
                    break
                else:
                    if isInt and "." in val:
                        isInt = False

            if isInt:
                spec.append(int)
            elif isFloat:
                spec.append(float)
            else:
                spec.append(str)

        return spec

    def _read(self) -> list | None:
        if self.pos < len(self):
            row = self.data[self.pos]
            self.pos += 1
            return row
        else:
            return None



"""Unit test for util.py"""
import geopandas as gpd
import numpy as np
import pytest

from ... import examples
from ...io.fileio import FileIO
from .. import util
from ..contiguity import Queen, Rook
from ..distance import KNN, DistanceBand
from ..util import fuzzy_contiguity, lat2W, nonplanar_neighbors
from ..weights import WSP, W


class Testutil:
    def setup_method(self):
        self.w = Rook.from_shapefile(examples.get_path("10740.shp"))

        self.rio = examples.load_example("Rio Grande do Sul")

    def test_lat2_w(self):
        w9 = lat2W(3, 3)
        assert w9.pct_nonzero == 29.62962962962963
        assert w9[0] == {1: 1.0, 3: 1.0}
        assert w9[3] == {0: 1.0, 4: 1.0, 6: 1.0}

    def test_lat2_sw(self):
        w9 = util.lat2SW(3, 3)
        rows, cols = w9.shape
        n = rows * cols
        assert w9.nnz == 24
        pct_nonzero = w9.nnz / float(n)
        assert pct_nonzero == 0.29629629629629628
        data = w9.todense().tolist()
        assert data[0] == [0, 1, 0, 1, 0, 0, 0, 0, 0]
        assert data[1] == [1, 0, 1, 0, 1, 0, 0, 0, 0]
        assert data[2] == [0, 1, 0, 0, 0, 1, 0, 0, 0]
        assert data[3] == [1, 0, 0, 0, 1, 0, 1, 0, 0]
        assert data[4] == [0, 1, 0, 1, 0, 1, 0, 1, 0]
        assert data[5] == [0, 0, 1, 0, 1, 0, 0, 0, 1]
        assert data[6] == [0, 0, 0, 1, 0, 0, 0, 1, 0]
        assert data[7] == [0, 0, 0, 0, 1, 0, 1, 0, 1]
        assert data[8] == [0, 0, 0, 0, 0, 1, 0, 1, 0]

    def test_block_weights(self):
        regimes = np.ones(25)
        regimes[list(range(10, 20))] = 2
        regimes[list(range(21, 25))] = 3
        regimes = np.array(
            [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                2.0,
                1.0,
                3.0,
                3.0,
                3.0,
                3.0,
            ]
        )
        w = util.block_weights(regimes)
        ww0 = [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        assert w.weights[0] == ww0
        wn0 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 20]
        assert w.neighbors[0] == wn0
        regimes = ["n", "n", "s", "s", "e", "e", "w", "w", "e"]
        w = util.block_weights(regimes)
        wn = {
            0: [1],
            1: [0],
            2: [3],
            3: [2],
            4: [5, 8],
            5: [4, 8],
            6: [7],
            7: [6],
            8: [4, 5],
        }
        assert w.neighbors == wn
        ids = ["id-%i" % i for i in range(len(regimes))]
        w = util.block_weights(regimes, ids=np.array(ids))
        w0 = {"id-1": 1.0}
        assert w["id-0"] == w0
        w = util.block_weights(regimes, ids=ids)
        w0 = {"id-1": 1.0}
        assert w["id-0"] == w0

    def test_comb(self):
        x = list(range(4))
        l_ = []
        for i in util.comb(x, 2):
            l_.append(i)
        lo = [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3]]
        assert l_ == lo

    def test_order(self):
        w3 = util.order(self.w, kmax=3)
        w3105 = [1, -1, 1, 2, 1]
        assert w3105 == w3[1][0:5]

    def test_higher_order(self):
        w10 = lat2W(10, 10)
        w10_2 = util.higher_order(w10, 2)
        w10_20 = {2: 1.0, 11: 1.0, 20: 1.0}
        assert w10_20 == w10_2[0]
        w5 = lat2W()
        w50 = {1: 1.0, 5: 1.0}
        assert w50 == w5[0]
        w51 = {0: 1.0, 2: 1.0, 6: 1.0}
        assert w51 == w5[1]
        w5_2 = util.higher_order(w5, 2)
        w5_20 = {2: 1.0, 10: 1.0, 6: 1.0}
        assert w5_20 == w5_2[0]

    def test_higher_order_sp(self):
        w10 = lat2W(10, 10)
        w10_3 = util.higher_order_sp(w10, 3)
        w10_30 = {30: 1.0, 21: 1.0, 12: 1.0, 3: 1.0}
        assert w10_30 == w10_3[0]
        w10_3 = util.higher_order_sp(w10, 3, lower_order=True)
        w10_30 = {
            20: 1.0,
            30: 1.0,
            21: 1.0,
            10: 1.0,
            1: 1.0,
            11: 1.0,
            2: 1.0,
            12: 1.0,
            3: 1.0,
        }
        assert w10_30 == w10_3[0]

    def test_higher_order_classes(self):
        wdb = DistanceBand.from_shapefile(examples.get_path("baltim.shp"), 34)
        wknn = KNN.from_shapefile(examples.get_path("baltim.shp"), 10)
        wrook = Rook.from_shapefile(examples.get_path("columbus.shp"))
        wqueen = Queen.from_shapefile(examples.get_path("columbus.shp"))
        wsparse = wqueen.sparse
        ww = W(wknn.neighbors, wknn.weights)
        util.higher_order(wdb, 2)
        util.higher_order(wknn, 3)
        util.higher_order(wrook, 4)
        util.higher_order(wqueen, 5)
        util.higher_order(wsparse, 2)
        util.higher_order(ww, 2)
        ww.transform = "r"
        _ = wrook.sparse
        util.higher_order(wsparse, 2)
        with pytest.raises(ValueError):
            util.higher_order(ww, 3)

    def test_shimbel(self):
        w5 = lat2W()
        w5_shimbel = util.shimbel(w5)
        w5_shimbel024 = 8
        assert w5_shimbel024 == w5_shimbel[0][24]
        w5_shimbel004 = [-1, 1, 2, 3]
        assert w5_shimbel004 == w5_shimbel[0][0:4]

    def test_full(self):
        neighbors = {
            "first": ["second"],
            "second": ["first", "third"],
            "third": ["second"],
        }
        weights = {"first": [1], "second": [1, 1], "third": [1]}
        w = W(neighbors, weights)
        wf, ids = util.full(w)
        wfo = np.array([[0.0, 1.0, 0.0], [1.0, 0.0, 1.0], [0.0, 1.0, 0.0]])
        np.testing.assert_array_almost_equal(wfo, wf, decimal=8)
        idso = ["first", "second", "third"]
        assert idso == ids

    def test_full2_w(self):
        a = np.zeros((4, 4))
        for i in range(len(a)):
            for j in range(len(a[i])):
                if i != j:
                    a[i, j] = np.random.random(1)[0]
        w = util.full2W(a)
        np.testing.assert_array_equal(w.full()[0], a)
        ids = ["myID0", "myID1", "myID2", "myID3"]
        w = util.full2W(a, ids=ids)
        np.testing.assert_array_equal(w.full()[0], a)

    def test_wsp2_w(self):
        sp = util.lat2SW(2, 5)
        wsp = WSP(sp)
        w = util.WSP2W(wsp)
        assert w.n == 10
        assert w[0] == {1: 1, 5: 1}
        for weights in w.weights.values():
            assert isinstance(weights, list)
        w = FileIO(examples.get_path("sids2.gal"), "r").read()
        wsp = WSP(w.sparse, w.id_order)
        w = util.WSP2W(wsp)
        assert w.n == 100
        assert w["37135"] == {
            "37001": 1.0,
            "37033": 1.0,
            "37037": 1.0,
            "37063": 1.0,
            "37145": 1.0,
        }

    def test_fill_diagonal(self):
        w1 = util.fill_diagonal(self.w)
        r1 = {0: 1.0, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        assert w1[0] == r1
        w1 = util.fill_diagonal(self.w, 20)
        r1 = {0: 20, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        assert w1[0] == r1
        diag = np.arange(100, 100 + self.w.n)
        w1 = util.fill_diagonal(self.w, diag)
        r1 = {0: 100, 1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}
        assert w1[0] == r1

    def test_remap_ids(self):
        w = lat2W(3, 2)
        wid_order = [0, 1, 2, 3, 4, 5]
        assert wid_order == w.id_order
        wneighbors0 = [2, 1]
        assert wneighbors0 == w.neighbors[0]
        old_to_new = {0: "a", 1: "b", 2: "c", 3: "d", 4: "e", 5: "f"}
        w_new = util.remap_ids(w, old_to_new)
        w_newid_order = ["a", "b", "c", "d", "e", "f"]
        assert w_newid_order == w_new.id_order
        w_newdneighborsa = ["c", "b"]
        assert w_newdneighborsa == w_new.neighbors["a"]

    def test_get_ids_shp(self):
        polyids = util.get_ids(examples.get_path("columbus.shp"), "POLYID")
        polyids5 = [1, 2, 3, 4, 5]
        assert polyids5 == polyids[:5]

    def test_get_ids_gdf(self):
        gdf = gpd.read_file(examples.get_path("columbus.shp"))
        polyids = util.get_ids(gdf, "POLYID")
        polyids5 = [1, 2, 3, 4, 5]
        assert polyids5 == polyids[:5]

    def test_get_points_array_from_shapefile(self):
        xy = util.get_points_array_from_shapefile(examples.get_path("juvenile.shp"))
        xy3 = np.array([[94.0, 93.0], [80.0, 95.0], [79.0, 90.0]])
        np.testing.assert_array_almost_equal(xy3, xy[:3], decimal=8)
        xy = util.get_points_array_from_shapefile(examples.get_path("columbus.shp"))
        xy3 = np.array(
            [
                [8.82721847, 14.36907602],
                [8.33265837, 14.03162401],
                [9.01226541, 13.81971908],
            ]
        )
        np.testing.assert_array_almost_equal(xy3, xy[:3], decimal=8)

    def test_min_threshold_distance(self):
        x, y = np.indices((5, 5))
        x.shape = (25, 1)
        y.shape = (25, 1)
        data = np.hstack([x, y])
        mint = 1.0
        assert mint == util.min_threshold_distance(data)

    def test_attach_islands(self):
        w = Rook.from_shapefile(examples.get_path("10740.shp"))
        w_knn1 = KNN.from_shapefile(examples.get_path("10740.shp"), k=1)
        w_attach = util.attach_islands(w, w_knn1)
        assert w_attach.islands == []
        assert w_attach[w.islands[0]] == {166: 1.0}

    def test_nonplanar_neighbors(self):
        df = gpd.read_file(examples.get_path("map_RS_BR.shp"))
        w = Queen.from_dataframe(df)
        assert w.islands == [
            0,
            4,
            23,
            27,
            80,
            94,
            101,
            107,
            109,
            119,
            122,
            139,
            169,
            175,
            223,
            239,
            247,
            253,
            254,
            255,
            256,
            261,
            276,
            291,
            294,
            303,
            321,
            357,
            374,
        ]
        wnp = nonplanar_neighbors(w, df)
        assert wnp.islands == []
        assert w.neighbors[0] == []
        assert wnp.neighbors[0] == [23, 59, 152, 239]
        assert wnp.neighbors[23] == [0, 45, 59, 107, 152, 185, 246]

    def test_fuzzy_contiguity(self):
        rs = examples.get_path("map_RS_BR.shp")
        rs_df = gpd.read_file(rs)
        wf = fuzzy_contiguity(rs_df)
        assert wf.islands == []
        assert set(wf.neighbors[0]) == {239, 59, 152, 23}
        buff = fuzzy_contiguity(rs_df, buffering=True, buffer=0.2)
        assert set(buff.neighbors[0]) == {175, 119, 239, 59, 152, 246, 23, 107}
        rs_index = rs_df.set_index("NM_MUNICIP")
        index_w = fuzzy_contiguity(rs_index)
        assert set(index_w.neighbors["TAVARES"]) == {"SÃO JOSÉ DO NORTE", "MOSTARDAS"}
        wf_pred = fuzzy_contiguity(rs_df, predicate="touches")
        assert set(wf_pred.neighbors[0]) == set()
        assert set(wf_pred.neighbors[1]) == {142, 82, 197, 285, 386, 350}



"""GeoDa Text File Reader Unit Tests"""

import pytest

from .... import examples as pysal_examples
from ..geoda_txt import GeoDaTxtReader


class TesttestGeoDaTxtReader:
    def setup_method(self):
        test_file = pysal_examples.get_path("stl_hom.txt")
        self.obj = GeoDaTxtReader(test_file, "r")

    def test___init__(self):
        assert self.obj.header == ["FIPSNO", "HR8488", "HR8893", "HC8488"]

    def test___len__(self):
        expected = 78
        assert expected == len(self.obj)

    def test_close(self):
        f = self.obj
        f.close()
        pytest.raises(ValueError, f.read)



# ruff: noqa: N802, N803

__all__ = ["KNN", "Kernel", "DistanceBand"]
__author__ = "Sergio J. Rey <srey@asu.edu>, Levi John Wolf <levi.john.wolf@gmail.com>"


import copy
from warnings import warn

import numpy as np
import scipy.sparse as sp
from scipy.spatial import distance_matrix

from ..cg.kdtree import KDTree
from .util import (
    WSP2W,
    get_ids,
    get_points_array,
    get_points_array_from_shapefile,
    isKDTree,
)
from .weights import WSP, W


def knnW(data, k=2, p=2, ids=None, radius=None, distance_metric="euclidean"):
    """
    This is deprecated. Use the pysal.weights.KNN class instead.
    """
    # warn('This function is deprecated. Please use pysal.weights.KNN', UserWarning)
    return KNN(data, k=k, p=p, ids=ids, radius=radius, distance_metric=distance_metric)


class KNN(W):
    """
    Creates nearest neighbor weights matrix based on k nearest
    neighbors.

    Parameters
    ----------
    kdtree      : object
                  PySAL KDTree or ArcKDTree where KDtree.data is array (n,k)
                  n observations on k characteristics used to measure
                  distances between the n objects
    k           : int
                  number of nearest neighbors
    p           : float
                  Minkowski p-norm distance metric parameter:
                  1<=p<=infinity
                  2: Euclidean distance
                  1: Manhattan distance
                  Ignored if the KDTree is an ArcKDTree
    ids         : list
                  identifiers to attach to each observation

    Returns
    -------
    w         : W
                instance
                Weights object with binary weights

    Examples
    --------
    >>> import libpysal
    >>> import numpy as np
    >>> points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    >>> kd = libpysal.cg.KDTree(np.array(points))
    >>> wnn2 = libpysal.weights.KNN(kd, 2)
    >>> [1,3] == wnn2.neighbors[0]
    True
    >>> wnn2 = KNN(kd,2)
    >>> wnn2[0]
    {1: 1.0, 3: 1.0}
    >>> wnn2[1]
    {0: 1.0, 3: 1.0}

    now with 1 rather than 0 offset

    >>> wnn2 = libpysal.weights.KNN(kd, 2, ids=range(1,7))
    >>> wnn2[1]
    {2: 1.0, 4: 1.0}
    >>> wnn2[2]
    {1: 1.0, 4: 1.0}
    >>> 0 in wnn2.neighbors
    False

    Notes
    -----
    Ties between neighbors of equal distance are arbitrarily broken.

    Further, if many points occupy the same spatial location (i.e. observations are
    coincident), then you may need to increase k for those observations to
    acquire neighbors at different spatial locations. For example, if five
    points are coincident, then their four nearest neighbors will all
    occupy the same spatial location; only the fifth nearest neighbor will
    result in those coincident points becoming connected to the graph as a
    whole.

    Solutions to this problem include jittering the points (by adding
    a small random value to each observation's location) or by adding
    higher-k neighbors only to the coincident points, using the
    weights.w_sets.w_union() function.

    See Also
    --------
    :class:`libpysal.weights.weights.W`
    """

    def __init__(
        self,
        data,
        k=2,
        p=2,
        ids=None,
        radius=None,
        distance_metric="euclidean",
        **kwargs,
    ):
        if radius is not None:
            distance_metric = "arc"
        if isKDTree(data):
            self.kdtree = data
            self.data = self.kdtree.data
        else:
            self.kdtree = KDTree(data, radius=radius, distance_metric=distance_metric)
            self.data = self.kdtree.data
        self.k = k
        self.p = p

        # these are both n x k+1
        distances, indices = self.kdtree.query(self.data, k=k + 1, p=p)
        full_indices = np.arange(self.kdtree.n)

        # if an element in the indices matrix is equal to the corresponding
        # index for that row, we want to mask that site from its neighbors
        not_self_mask = indices != full_indices.reshape(-1, 1)
        # if there are *too many duplicates per site*, then we may get some
        # rows where the site index is not in the set of k+1 neighbors
        # So, we need to know where these sites are
        has_one_too_many = not_self_mask.sum(axis=1) == (k + 1)
        # if a site has k+1 neighbors, drop its k+1th neighbor
        not_self_mask[has_one_too_many, -1] &= False
        not_self_indices = indices[not_self_mask].reshape(self.kdtree.n, -1)

        if ids is None:
            ids = list(full_indices)
            named_indices = not_self_indices
        else:
            named_indices = np.asarray(ids)[not_self_indices]
        neighbors = {
            idx: list(indices) for idx, indices in zip(ids, named_indices, strict=True)
        }

        W.__init__(self, neighbors, id_order=ids, **kwargs)

    @classmethod
    def from_shapefile(cls, filepath, *args, **kwargs):
        """
        Nearest neighbor weights from a shapefile.

        Parameters
        ----------
        data       : string
                     shapefile containing attribute data.
        k          : int
                     number of nearest neighbors
        p          : float
                     Minkowski p-norm distance metric parameter:
                     1<=p<=infinity
                     2: Euclidean distance
                     1: Manhattan distance
        ids        : list
                     identifiers to attach to each observation
        radius     : float
                     If supplied arc_distances will be calculated
                     based on the given radius. p will be ignored.

        Returns
        -------
        w         : KNN
                    instance; Weights object with binary weights.

        Examples
        --------
        Polygon shapefile
        >>> import libpysal
        >>> from libpysal.weights import KNN
        >>> wc=KNN.from_shapefile(libpysal.examples.get_path("columbus.shp"))
        >>> "%.4f"%wc.pct_nonzero
        '4.0816'
        >>> set([2,1]) == set(wc.neighbors[0])
        True
        >>> wc3=KNN.from_shapefile(libpysal.examples.get_path("columbus.shp"),k=3)
        >>> set(wc3.neighbors[0]) == set([2,1,3])
        True
        >>> set(wc3.neighbors[2]) == set([4,3,0])
        True

        Point shapefile

        >>> w=KNN.from_shapefile(libpysal.examples.get_path("juvenile.shp"))
        >>> w.pct_nonzero
        1.1904761904761905
        >>> w1=KNN.from_shapefile(libpysal.examples.get_path("juvenile.shp"),k=1)
        >>> "%.3f"%w1.pct_nonzero
        '0.595'

        Notes
        -----
        Ties between neighbors of equal distance are arbitrarily broken.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        return cls(get_points_array_from_shapefile(filepath), *args, **kwargs)

    @classmethod
    def from_array(cls, array, *args, **kwargs):
        """
        Creates nearest neighbor weights matrix based on k nearest
        neighbors.

        Parameters
        ----------
        array       : np.ndarray
                      (n, k) array representing n observations on
                      k characteristics used to measure distances
                      between the n objects
        **kwargs    : keyword arguments, see Rook

        Returns
        -------
        w         : W
                    instance
                    Weights object with binary weights

        Examples
        --------
        >>> from libpysal.weights import KNN
        >>> points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
        >>> wnn2 = KNN.from_array(points, 2)
        >>> [1,3] == wnn2.neighbors[0]
        True
        >>> wnn2 = KNN.from_array(points,2)
        >>> wnn2[0]
        {1: 1.0, 3: 1.0}
        >>> wnn2[1]
        {0: 1.0, 3: 1.0}

        now with 1 rather than 0 offset

        >>> wnn2 = KNN.from_array(points, 2, ids=range(1,7))
        >>> wnn2[1]
        {2: 1.0, 4: 1.0}
        >>> wnn2[2]
        {1: 1.0, 4: 1.0}
        >>> 0 in wnn2.neighbors
        False

        Notes
        -----
        Ties between neighbors of equal distance are arbitrarily broken.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        return cls(array, *args, **kwargs)

    @classmethod
    def from_dataframe(
        cls, df, geom_col=None, ids=None, use_index=True, *args, **kwargs
    ):
        """
        Make KNN weights from a dataframe.

        Parameters
        ----------
        df      :   pandas.dataframe
                    a dataframe with a geometry column that can be used to
                    construct a W object
        geom_col :  string
                    the name of the column in `df` that contains the
                    geometries. Defaults to active geometry column.
        ids     :   list-like, string
                    a list-like of ids to use to index the spatial weights object or
                    the name of the column to use as IDs. If nothing is
                    provided, the dataframe index is used if `use_index=True` or
                    a positional index is used if `use_index=False`.
                    Order of the resulting W is not respected from this list.
        use_index   : bool
                    use index of `df` as `ids` to index the spatial weights object.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        if geom_col is None:
            geom_col = df.geometry.name
        pts = get_points_array(df[geom_col])
        if ids is None and use_index:
            ids = df.index.tolist()
        elif isinstance(ids, str):
            ids = df[ids].tolist()
        return cls(pts, *args, ids=ids, **kwargs)

    def reweight(self, k=None, p=None, new_data=None, new_ids=None, inplace=True):
        """
        Redo K-Nearest Neighbor weights construction using given parameters

        Parameters
        ----------
        new_data    : np.ndarray
                      an array containing additional data to use in the KNN
                      weight
        new_ids     : list
                      a list aligned with new_data that provides the ids for
                      each new observation
        inplace     : bool
                      a flag denoting whether to modify the KNN object
                      in place or to return a new KNN object
        k           : int
                      number of nearest neighbors
        p           : float
                      Minkowski p-norm distance metric parameter:
                      1<=p<=infinity
                      2: Euclidean distance
                      1: Manhattan distance
                      Ignored if the KDTree is an ArcKDTree

        Returns
        -------
        A copy of the object using the new parameterization, or None if the
        object is reweighted in place.
        """

        if new_data is not None:
            new_data = np.asarray(new_data).reshape(-1, 2)
            data = np.vstack((self.data, new_data)).reshape(-1, 2)
            if new_ids is not None:
                ids = copy.deepcopy(self.id_order)
                ids.extend(list(new_ids))
            else:
                ids = list(range(data.shape[0]))
        elif (new_data is None) and (new_ids is None):
            # If not, we can use the same kdtree we have
            data = self.kdtree
            ids = self.id_order
        elif (new_data is None) and (new_ids is not None):
            warn("Remapping ids must be done using w.remap_ids", stacklevel=2)
        if k is None:
            k = self.k
        if p is None:
            p = self.p
        if inplace:
            self._reset()
            self.__init__(data, ids=ids, k=k, p=p)
        else:
            return KNN(data, ids=ids, k=k, p=p)


class Kernel(W):
    """
    Spatial weights based on kernel functions.

    Parameters
    ----------
    data        : array
                  (n,k) or KDTree where KDtree.data is array (n,k)
                  n observations on k characteristics used to measure
                  distances between the n objects
    bandwidth   : float
                  or array-like (optional)
                  the bandwidth :math:`h_i` for the kernel.
    fixed       : binary
                  If true then :math:`h_i=h \\forall i`. If false then
                  bandwidth is adaptive across observations.
    k           : int
                  the number of nearest neighbors to use for determining
                  bandwidth. For fixed bandwidth, :math:`h_i=max(dknn) \\forall i`
                  where :math:`dknn` is a vector of k-nearest neighbor
                  distances (the distance to the kth nearest neighbor for each
                  observation).  For adaptive bandwidths, :math:`h_i=dknn_i`
    diagonal    : boolean
                  If true, set diagonal weights = 1.0, if false (default),
                  diagonals weights are set to value according to kernel
                  function.
    function    : {'triangular','uniform','quadratic','quartic','gaussian'}
                  kernel function defined as follows with

                  .. math::

                      z_{i,j} = d_{i,j}/h_i

                  triangular

                  .. math::

                      K(z) = (1 - |z|) \\ if |z| \\le 1

                  uniform

                  .. math::

                      K(z) = 1/2 \\ if |z| \\le 1

                  quadratic

                  .. math::

                      K(z) = (3/4)(1-z^2) \\ if |z| \\le 1

                  quartic

                  .. math::

                      K(z) = (15/16)(1-z^2)^2 \\ if |z| \\le 1

                  gaussian

                  .. math::

                      K(z) = (2\\pi)^{(-1/2)} exp(-z^2 / 2)
    eps         : float
                  adjustment to ensure knn distance range is closed on the
                  knnth observations

    Attributes
    ----------
    weights : dict
              Dictionary keyed by id with a list of weights for each neighbor
    neighbors : dict
                of lists of neighbors keyed by observation id
    bandwidth : array
                array of bandwidths

    Examples
    --------
    >>> from libpysal.weights import Kernel
    >>> points=[(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    >>> kw=Kernel(points)
    >>> kw.weights[0]
    [1.0, 0.500000049999995, 0.4409830615267465]
    >>> kw.neighbors[0]
    [0, 1, 3]
    >>> kw.bandwidth
    array([[20.000002],
           [20.000002],
           [20.000002],
           [20.000002],
           [20.000002],
           [20.000002]])
    >>> kw15=Kernel(points,bandwidth=15.0)
    >>> kw15[0]
    {0: 1.0, 1: 0.33333333333333337, 3: 0.2546440075000701}
    >>> kw15.neighbors[0]
    [0, 1, 3]
    >>> kw15.bandwidth
    array([[15.],
           [15.],
           [15.],
           [15.],
           [15.],
           [15.]])

    Adaptive bandwidths user specified

    >>> bw=[25.0,15.0,25.0,16.0,14.5,25.0]
    >>> kwa=Kernel(points,bandwidth=bw)
    >>> kwa.weights[0]
    [1.0, 0.6, 0.552786404500042, 0.10557280900008403]
    >>> kwa.neighbors[0]
    [0, 1, 3, 4]
    >>> kwa.bandwidth
    array([[25. ],
           [15. ],
           [25. ],
           [16. ],
           [14.5],
           [25. ]])

    Endogenous adaptive bandwidths

    >>> kwea=Kernel(points,fixed=False)
    >>> kwea.weights[0]
    [1.0, 0.10557289844279438, 9.99999900663795e-08]
    >>> kwea.neighbors[0]
    [0, 1, 3]
    >>> kwea.bandwidth
    array([[11.18034101],
           [11.18034101],
           [20.000002  ],
           [11.18034101],
           [14.14213704],
           [18.02775818]])

    Endogenous adaptive bandwidths with Gaussian kernel

    >>> kweag=Kernel(points,fixed=False,function='gaussian')
    >>> kweag.weights[0]
    [0.3989422804014327, 0.2674190291577696, 0.2419707487162134]
    >>> kweag.bandwidth
    array([[11.18034101],
           [11.18034101],
           [20.000002  ],
           [11.18034101],
           [14.14213704],
           [18.02775818]])

    Diagonals to 1.0

    >>> kq = Kernel(points,function='gaussian')
    >>> kq.weights
    {0: [0.3989422804014327, 0.35206533556593145, 0.3412334260702758], 1: [0.35206533556593145, 0.3989422804014327, 0.2419707487162134, 0.3412334260702758, 0.31069657591175387], 2: [0.2419707487162134, 0.3989422804014327, 0.31069657591175387], 3: [0.3412334260702758, 0.3412334260702758, 0.3989422804014327, 0.3011374490937829, 0.26575287272131043], 4: [0.31069657591175387, 0.31069657591175387, 0.3011374490937829, 0.3989422804014327, 0.35206533556593145], 5: [0.26575287272131043, 0.35206533556593145, 0.3989422804014327]}
    >>> kqd = Kernel(points, function='gaussian', diagonal=True)
    >>> kqd.weights
    {0: [1.0, 0.35206533556593145, 0.3412334260702758], 1: [0.35206533556593145, 1.0, 0.2419707487162134, 0.3412334260702758, 0.31069657591175387], 2: [0.2419707487162134, 1.0, 0.31069657591175387], 3: [0.3412334260702758, 0.3412334260702758, 1.0, 0.3011374490937829, 0.26575287272131043], 4: [0.31069657591175387, 0.31069657591175387, 0.3011374490937829, 1.0, 0.35206533556593145], 5: [0.26575287272131043, 0.35206533556593145, 1.0]}

    """  # noqa: E501

    def __init__(
        self,
        data,
        bandwidth=None,
        fixed=True,
        k=2,
        function="triangular",
        eps=1.0000001,
        ids=None,
        diagonal=False,
        distance_metric="euclidean",
        radius=None,
        **kwargs,
    ):
        if radius is not None:
            distance_metric = "arc"
        if isKDTree(data):
            self.kdtree = data
            self.data = self.kdtree.data
            data = self.data
        else:
            self.kdtree = KDTree(data, distance_metric=distance_metric, radius=radius)
            self.data = self.kdtree.data
        self.k = k + 1
        self.function = function.lower()
        self.fixed = fixed
        self.eps = eps
        if bandwidth:
            try:
                bandwidth = np.array(bandwidth)
                bandwidth.shape = (len(bandwidth), 1)
            except:  # noqa: E722
                bandwidth = np.ones((len(data), 1), "float") * bandwidth
            self.bandwidth = bandwidth
        else:
            self._set_bw()

        self._eval_kernel()
        neighbors, weights = self._k_to_W(ids)
        if diagonal:
            for i in neighbors:
                weights[i][neighbors[i].index(i)] = 1.0
        W.__init__(self, neighbors, weights, ids, **kwargs)

    @classmethod
    def from_shapefile(cls, filepath, idVariable=None, **kwargs):
        """
        Kernel based weights from shapefile

        Parameters
        ----------
        shapefile   : string
                      shapefile name with shp suffix
        idVariable  : string
                      name of column in shapefile's DBF to use for ids

        Returns
        -------
        Kernel Weights Object

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        points = get_points_array_from_shapefile(filepath)
        ids = get_ids(filepath, idVariable) if idVariable is not None else None
        return cls.from_array(points, ids=ids, **kwargs)

    @classmethod
    def from_array(cls, array, **kwargs):
        """
        Construct a Kernel weights from an array. Supports all the same options
        as :class:`libpysal.weights.Kernel`

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        return cls(array, **kwargs)

    @classmethod
    def from_dataframe(cls, df, geom_col=None, ids=None, use_index=True, **kwargs):
        """
        Make Kernel weights from a dataframe.

        Parameters
        ----------
        df      :   pandas.dataframe
                    a dataframe with a geometry column that can be used to
                    construct a W object
        geom_col :  string
                    the name of the column in `df` that contains the
                    geometries. Defaults to active geometry column.
        ids     :   list-like, string
                    a list-like of ids to use to index the spatial weights object or
                    the name of the column to use as IDs. If nothing is
                    provided, the dataframe index is used if `use_index=True` or
                    a positional index is used if `use_index=False`.
                    Order of the resulting W is not respected from this list.
        use_index   : bool
                    use index of `df` as `ids` to index the spatial weights object.

        See Also
        --------
        :class:`libpysal.weights.weights.W`
        """
        if geom_col is None:
            geom_col = df.geometry.name
        pts = get_points_array(df[geom_col])
        if ids is None and use_index:
            ids = df.index.tolist()
        elif isinstance(ids, str):
            ids = df[ids].tolist()
        return cls(pts, ids=ids, **kwargs)

    def _k_to_W(self, ids=None):
        allneighbors = {}
        weights = {}
        ids = np.array(ids) if ids else np.arange(len(self.data))
        for i, _ in enumerate(self.kernel):
            if len(self.neigh[i]) == 0:
                allneighbors[ids[i]] = []
                weights[ids[i]] = []
            else:
                allneighbors[ids[i]] = list(ids[self.neigh[i]])
                weights[ids[i]] = self.kernel[i].tolist()
        return allneighbors, weights

    def _set_bw(self):
        dmat, neigh = self.kdtree.query(self.data, k=self.k)
        if self.fixed:
            # use max knn distance as bandwidth
            bandwidth = dmat.max() * self.eps
            n = len(dmat)
            self.bandwidth = np.ones((n, 1), "float") * bandwidth
        else:
            # use local max knn distance
            self.bandwidth = dmat.max(axis=1) * self.eps
            self.bandwidth.shape = (self.bandwidth.size, 1)
            # identify knn neighbors for each point
            nnq = self.kdtree.query(self.data, k=self.k)
            self.neigh = nnq[1]

    def _eval_kernel(self):
        # get points within bandwidth distance of each point
        if not hasattr(self, "neigh"):
            kdtq = self.kdtree.query_ball_point
            neighbors = [
                kdtq(self.data[i], r=bwi[0]) for i, bwi in enumerate(self.bandwidth)
            ]
            self.neigh = neighbors
        # get distances for neighbors
        bw = self.bandwidth

        kdtq = self.kdtree.query
        z = []
        for i, nids in enumerate(self.neigh):
            di, ni = kdtq(self.data[i], k=len(nids))
            if not isinstance(di, np.ndarray):
                di = np.asarray([di] * len(nids))
                ni = np.asarray([ni] * len(nids))
            zi = (
                np.array([dict(list(zip(ni, di, strict=True)))[nid] for nid in nids])
                / bw[i]
            )
            z.append(zi)
        zs = z
        # functions follow Anselin and Rey (2010) table 5.4
        if self.function == "triangular":
            self.kernel = [1 - zi for zi in zs]
        elif self.function == "uniform":
            self.kernel = [np.ones(zi.shape) * 0.5 for zi in zs]
        elif self.function == "quadratic":
            self.kernel = [(3.0 / 4) * (1 - zi**2) for zi in zs]
        elif self.function == "quartic":
            self.kernel = [(15.0 / 16) * (1 - zi**2) ** 2 for zi in zs]
        elif self.function == "gaussian":
            c = np.pi * 2
            c = c ** (-0.5)
            self.kernel = [c * np.exp(-(zi**2) / 2.0) for zi in zs]
        else:
            print(("Unsupported kernel function", self.function))


class DistanceBand(W):
    """
    Spatial weights based on distance band.

    Parameters
    ----------
    data        : array
                  (n,k) or KDTree where KDtree.data is array (n,k)
                  n observations on k characteristics used to measure
                  distances between the n objects
    threshold  : float
                 distance band
    p          : float
                 DEPRECATED: use `distance_metric`
                 Minkowski p-norm distance metric parameter:
                 1<=p<=infinity
                 2: Euclidean distance
                 1: Manhattan distance
    binary     : boolean
                 If true w_{ij}=1 if d_{i,j}<=threshold, otherwise w_{i,j}=0
                 If false wij=dij^{alpha}
    alpha      : float
                 distance decay parameter for weight (default -1.0)
                 if alpha is positive the weights will not decline with
                 distance. If binary is True, alpha is ignored
    ids         : list
                  values to use for keys of the neighbors and weights dicts
    build_sp    : boolean
                  DEPRECATED
                  True to build sparse distance matrix and false to build dense
                  distance matrix; significant speed gains may be obtained
                  dending on the sparsity of the of distance_matrix and
                  threshold that is applied
    silent      : boolean
                  By default libpysal will print a warning if the
                  dataset contains any disconnected observations or
                  islands. To silence this warning set this
                  parameter to True.

    Attributes
    ----------
    weights : dict
              of neighbor weights keyed by observation id

    neighbors : dict
                of neighbors keyed by observation id

    Examples
    --------
    >>> import libpysal
    >>> points=[(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    >>> wcheck = libpysal.weights.W(
    ...     {0: [1, 3], 1: [0, 3], 2: [], 3: [0, 1], 4: [5], 5: [4]}
    ... )

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> w=libpysal.weights.DistanceBand(points,threshold=11.2)

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> libpysal.weights.util.neighbor_equality(w, wcheck)
    True
    >>> w=libpysal.weights.DistanceBand(points,threshold=14.2)
    >>> wcheck = libpysal.weights.W(
    ...     {0: [1, 3], 1: [0, 3, 4], 2: [4], 3: [1, 0], 4: [5, 2, 1], 5: [4]}
    ... )
    >>> libpysal.weights.util.neighbor_equality(w, wcheck)
    True

    inverse distance weights

    >>> w=libpysal.weights.DistanceBand(points,threshold=11.2,binary=False)

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> w.weights[0]
    [0.1, 0.08944271909999159]
    >>> w.neighbors[0].tolist()
    [1, 3]

    gravity weights

    >>> w=libpysal.weights.DistanceBand(points,threshold=11.2,binary=False,alpha=-2.)

    WARNING: there is one disconnected observation (no neighbors)
    Island id:  [2]
    >>> w.weights[0]
    [0.01, 0.007999999999999998]
    """

    def __init__(
        self,
        data,
        threshold,
        p=2,
        alpha=-1.0,
        binary=True,
        ids=None,
        build_sp=True,
        silence_warnings=False,
        distance_metric="euclidean",
        radius=None,
    ):
        """Casting to floats is a work around for a bug in scipy.spatial.
        See detail in pysal issue #126.
        """
        if ids is not None:
            ids = list(ids)
        if radius is not None:
            distance_metric = "arc"
        self.p = p
        self.threshold = threshold
        self.binary = binary
        self.alpha = alpha
        self.build_sp = build_sp
        self.silence_warnings = silence_warnings

        if isKDTree(data):
            self.kdtree = data
            self.data = self.kdtree.data
        else:
            if self.build_sp:
                try:
                    data = np.asarray(data)
                    if data.dtype.kind != "f":
                        data = data.astype(float)
                    self.kdtree = KDTree(
                        data, distance_metric=distance_metric, radius=radius
                    )
                    self.data = self.kdtree.data
                except:  # noqa: E722
                    raise ValueError("Could not make array from data") from None
            else:
                self.data = data
                self.kdtree = None

        self._band()
        neighbors, weights = self._distance_to_W(ids)
        W.__init__(
            self, neighbors, weights, ids, silence_warnings=self.silence_warnings
        )

    @classmethod
    def from_shapefile(cls, filepath, threshold, idVariable=None, **kwargs):
        """
        Distance-band based weights from shapefile

        Parameters
        ----------
        shapefile   : string
                      shapefile name with shp suffix
        idVariable  : string
                      name of column in shapefile's DBF to use for ids

        Returns
        -------
        Kernel Weights Object
        """
        points = get_points_array_from_shapefile(filepath)
        ids = get_ids(filepath, idVariable) if idVariable is not None else None
        return cls.from_array(points, threshold, ids=ids, **kwargs)

    @classmethod
    def from_array(cls, array, threshold, **kwargs):
        """
        Construct a DistanceBand weights from an array. Supports all the same options
        as :class:`libpysal.weights.DistanceBand`
        """
        return cls(array, threshold, **kwargs)

    @classmethod
    def from_dataframe(
        cls, df, threshold, geom_col=None, ids=None, use_index=True, **kwargs
    ):
        """
        Make DistanceBand weights from a dataframe.

        Parameters
        ----------
        df      :   pandas.dataframe
                    a dataframe with a geometry column that can be used to
                    construct a W object
        geom_col :  string
                    the name of the column in `df` that contains the
                    geometries. Defaults to active geometry column.
        ids     :   list-like, string
                    a list-like of ids to use to index the spatial weights object or
                    the name of the column to use as IDs. If nothing is
                    provided, the dataframe index is used if `use_index=True` or
                    a positional index is used if `use_index=False`.
                    Order of the resulting W is not respected from this list.
        use_index   : bool
                    use index of `df` as `ids` to index the spatial weights object.
        """
        if geom_col is None:
            geom_col = df.geometry.name
        pts = get_points_array(df[geom_col])
        if ids is None and use_index:
            ids = df.index.tolist()
        elif isinstance(ids, str):
            ids = df[ids].tolist()
        return cls(pts, threshold, ids=ids, **kwargs)

    def _band(self):
        """Find all pairs within threshold."""
        if self.build_sp:
            self.dmat = self.kdtree.sparse_distance_matrix(
                self.kdtree, max_distance=self.threshold, p=self.p
            ).tocsr()
        else:
            if str(self.kdtree).split(".")[-1][0:10] == "Arc_KDTree":
                raise TypeError(
                    "Unable to calculate dense arc distance matrix;"
                    ' parameter "build_sp" must be set to True for arc'
                    " distance type weight"
                )
            self.dmat = self._spdistance_matrix(self.data, self.data, self.threshold)

    def _distance_to_W(self, ids=None):
        if self.binary:
            self.dmat[self.dmat > 0] = 1
            self.dmat.eliminate_zeros()
            temp_w = WSP2W(
                WSP(self.dmat, id_order=ids), silence_warnings=self.silence_warnings
            )
            neighbors = temp_w.neighbors
            weight_keys = list(temp_w.weights.keys())
            weight_vals = list(temp_w.weights.values())
            weights = dict(
                list(zip(weight_keys, list(map(list, weight_vals)), strict=True))
            )
            return neighbors, weights
        else:
            weighted = self.dmat.power(self.alpha)
            weighted[weighted == np.inf] = 0
            weighted.eliminate_zeros()
            temp_w = WSP2W(
                WSP(weighted, id_order=ids), silence_warnings=self.silence_warnings
            )
            neighbors = temp_w.neighbors
            weight_keys = list(temp_w.weights.keys())
            weight_vals = list(temp_w.weights.values())
            weights = dict(
                list(zip(weight_keys, list(map(list, weight_vals)), strict=True))
            )
            return neighbors, weights

    def _spdistance_matrix(self, x, y, threshold=None):
        dist = distance_matrix(x, y)
        if threshold is not None:
            zeros = dist > threshold
            dist[zeros] = 0
        return sp.csr_matrix(dist)


def _test():
    import doctest

    # the following line could be used to define an alternative to the
    # '<BLANKLINE>' flag doctest.BLANKLINE_MARKER = 'something better than <BLANKLINE>'
    start_suppress = np.get_printoptions()["suppress"]
    np.set_printoptions(suppress=True)
    doctest.testmod()
    np.set_printoptions(suppress=start_suppress)


if __name__ == "__main__":
    _test()



""" Example Reader and Writer

These are working readers/writers that parse '.foo' and '.bar' files.

"""

# ruff: noqa: N812, SIM115

from .. import fileio as FileIO

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["TemplateWriter", "TemplateReaderWriter"]


# Always subclass FileIO
class TemplateWriter(FileIO.FileIO):
    # REQUIRED, List the formats this class supports.
    FORMATS = ["foo"]

    # REQUIRED, List the modes supported by this class.
    # One class can support both reading and writing.
    # For simplicity this class will only support one.
    # You could support custom modes, but these could be hard to document.
    MODES = ["w"]

    # Use ``.__init__`` to open any need file handlers
    def __init__(self, *args, **kwargs):
        # initialize the parent class...
        FileIO.__init__(self, *args, **kwargs)

        # this gives you:
        # self.dataPath == the connection string or path to file
        # self.mode == the mode the file should be opened in

        self.fileObj = open(self.dataPath, self.mode)

    # Writers must subclass ``.write()``
    def write(self, obj):
        """``.write`` method of the 'foobar' template

        Parameters
        ----------
        obj : str
            Some string.

        Raises
        ------
        TypeError
            Raised when a ``str`` is expected, but got another type.

        """

        # GOOD TO HAVE, this will prevent invalid operations on closed files.
        self._complain_ifclosed(self.closed)

        # It's up to the writer to understand the object, you should check
        # that object is of the type you expect and raise a TypeError is its now.
        # we will support writing string objects in this example,
        # all string are derived from basestring...
        if issubclass(type(obj), str):
            # Non-essential...
            def foobar(c):
                return c in "foobar"

            # e.g. 'foobara' == filter(foobar,'my little foobar example')
            result = list(filter(foobar, obj))

            # do the actual writing...
            self.fileObj.write(result + "\n")

            # REQUIRED, increment the internal pos pointer.
            self.pos += 1

        else:
            raise TypeError("Expected a string, got: %s." % (type(obj)))

    # default is to raise "NotImplementedError"
    def flush(self):
        self._complain_ifclosed(self.closed)
        self.fileObj.flush()

    # REQUIRED
    def close(self):
        self.fileObj.close()
        # clean up the parent class too....
        FileIO.close(self)


class TemplateReaderWriter(FileIO.FileIO):
    FORMATS = ["bar"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        FileIO.__init__(self, *args, **kwargs)
        self.fileObj = open(self.dataPath, self.mode)

    # Notice reading is a bit different

    def _filter(self, st):
        def foobar(c):
            return c in "foobar"

        # e.g. 'foobara' == filter(foobar,'my little foobar example')
        return list(filter(foobar, st))

    def _read(self):
        """The ``_read`` method should return only ONE object.

        Returns
        -------
        obj_plus_break : str
            only ONE object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        """

        line = self.fileObj.readline()
        obj = self._filter(line)

        # REQUIRED
        self.pos += 1
        if line:
            obj_plus_break = obj + "\n"
            return obj_plus_break
        else:
            # REQUIRED
            raise StopIteration

    def write(self, obj):
        """The ``.write`` method of the 'foobar' template, receives an ``obj``.

        Paramters
        ---------
        obj : str
            Some string.

        Raises
        ------
        TypeError
            Raised when a ``str`` is expected, but got another type.

        """
        self._complain_ifclosed(self.closed)
        if issubclass(type(obj), str):
            result = self._filter(obj)
            self.fileObj.write(result + "\n")
            self.pos += 1
        else:
            raise TypeError("Expected a string, got: %s" % (type(obj)))

    def flush(self):
        self._complain_ifclosed(self.closed)
        self.fileObj.flush()

    def close(self):
        self.fileObj.close()
        FileIO.close(self)


# if __name__ == "__main__":
#     "NOTE, by running OR importing this module"
#     "it's automatically added to the pysal fileIO registry."
#
#     pysal.open.check()  # noqa
#
#     lines = [
#         "This is an example of template FileIO classes",
#         "Each call to write expects a string object",
#         "that string is filtered and only letters 'f','o','b','a','r' are kept",
#         "these kept letters are written to the file",
#         "and a new line char is appends to each line",
#         "likewise the reader filters each line from a file.",
#     ]
#
#     f = pysal.open("test.foo", "w")  # noqa
#     for line in lines:
#         f.write(line)
#     f.close()
#
#     f = pysal.open("test.bar", "w")  # noqa
#     for line in lines:
#         f.write(line)
#     f.close()
#
#     f = pysal.open("test.bar", "r")  # noqa
#     s = "".join(f.read())
#     f.close()
#     print(s)
#
#     f = open("test.foo")
#     s2 = f.read()
#     f.close()
#     print(s == s2)



# ruff: noqa: N801, N802,N812, N815, SIM115

import os.path
from warnings import warn

from ...weights.weights import W
from .. import fileio as FileIO

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["GwtIO"]


class unique_filter:
    """(Util function) When a new instance is passed as an arugment to the
    builtin filter it will remove duplicate entries without changing the
    order of the list. Be sure to ceate a new instance everytime, unless
    you want a global filter.

    Examples
    --------

    >>> l = ['a', 'a', 'b', 'a', 'c', 'v', 'd', 'a', 'v', 'd']
    >>> list(filter(unique_filter(),l))
    ['a', 'b', 'c', 'v', 'd']

    """

    def __init__(self):
        self.exclude = set()

    def __call__(self, x) -> bool:
        if x in self.exclude:
            return False
        else:
            self.exclude.add(x)
            return True


class GwtIO(FileIO.FileIO):
    FORMATS = ["kwt", "gwt"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        self._varName = "Unknown"
        self._shpName = "Unknown"
        FileIO.FileIO.__init__(self, *args, **kwargs)
        self.file = open(self.dataPath, self.mode)

    def _set_varName(self, val):
        if issubclass(type(val), str):
            self._varName = val

    def _get_varName(self) -> str:
        return self._varName

    varName = property(fget=_get_varName, fset=_set_varName)

    def _set_shpName(self, val):
        if issubclass(type(val), str):
            self._shpName = val

    def _get_shpName(self) -> str:
        return self._shpName

    shpName = property(fget=_get_shpName, fset=_set_shpName)

    def read(self, n=-1):  # noqa: ARG002
        """

        Parameters
        ----------
        n : int
            Read at most ``n`` objects. Default is ``-1``.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        """

        self._complain_ifclosed(self.closed)

        w = self._read()

        return w

    def seek(self, pos):
        if pos == 0:
            self.file.seek(0)
            self.pos = 0

    def _readlines(self, id_type, ret_ids=False):
        """Reads the main body of gwt-like weights files into two dictionaries
        containing weights and neighbors. This code part is repeatedly used for
        many weight file formats. Header lines, however, are different from
        format to format. So, for code reusability, this part is separated out
        from the ``_read()`` function by Myunghwa Hwang.

        Parameters
        ----------
        id_type : type
            Cast IDs as this type.
        ret_ids : bool
            Return IDs (``True``). Default is ``False``.

        Returns
        -------
        weights : dict
            Dictionary of weight values.
        neighbors : dict
            Dictionary of neighbor ID values.
        ids : list
            List of ID values.

        """

        data = [row.strip().split() for row in self.file.readlines()]
        ids = list(filter(unique_filter(), [x[0] for x in data]))
        ids = list(map(id_type, ids))
        wn = {}

        # note: fromkeys is no good here, all keys end up sharing the say dict value
        for id_ in ids:
            wn[id_] = {}

        for i, j, v in data:
            i = id_type(i)
            j = id_type(j)
            wn[i][j] = float(v)
        weights = {}
        neighbors = {}

        for i in wn:
            weights[i] = list(wn[i].values())
            neighbors[i] = list(wn[i].keys())
        if ret_ids:
            return weights, neighbors, ids
        else:
            return weights, neighbors

    def _read(self):
        """Reads ``.gwt`` file.

        Returns
        -------
        w : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.

        Examples
        --------

        Type ``dir(f)`` at the interpreter to see what methods are supported.
        Open ``.gwt`` file and read it into a PySAL weights object.

        >>> import libpysal
        >>> f = libpysal.io.open(libpysal.examples.get_path('juvenile.gwt'), 'r').read()

        Get the number of observations from the header.

        >>> f.n
        168

        Get the mean number of neighbors.

        >>> f.mean_neighbors
        16.678571428571427

        Get neighbor distances for a single observation.

        >>> f[1]
        {2: 14.1421356}

        """

        if self.pos > 0:
            raise StopIteration

        flag, n, shp, id_var = self.file.readline().strip().split()
        self.shpName = shp
        self.varName = id_var
        id_order = None
        id_type = str

        try:
            base = os.path.split(self.dataPath)[0]
            dbf = os.path.join(base, self.shpName.replace(".shp", "") + ".dbf")
            if os.path.exists(dbf):
                db = FileIO.FileIO(dbf, "r")
                if id_var in db.header:
                    id_order = db.by_col(id_var)
                    id_type = type(id_order[0])
                else:
                    msg = "ID_VAR:'%s' was in in the DBF header, "
                    msg += "proceeding with unordered string IDs."
                    msg = msg % id_var
                    warn(msg, RuntimeWarning, stacklevel=2)
            else:
                msg = "DBF relating to GWT was not found, "
                msg += "proceeding with unordered string IDs."
                warn(msg, RuntimeWarning, stacklevel=2)
        except:  # noqa: E722
            msg = "Exception occurred will reading DBF, "
            msg += "proceeding with unordered string IDs."
            warn(msg, RuntimeWarning, stacklevel=2)

        self.flag = flag
        self.n = n
        self.shp = shp
        self.id_var = id_var

        if id_order is None:
            weights, neighbors, id_order = self._readlines(id_type, True)
        else:
            weights, neighbors = self._readlines(id_type)

        self.pos += 1
        w = W(neighbors, weights, id_order)
        # w.transform = 'b'

        # set meta data
        w._shpName = self.shpName
        w._varName = self.varName

        # msg = "Weights have been converted to binary. "
        # msg += "To retrieve original values use w.transform='o'"
        # warn(msg, RuntimeWarning)

        return w

    def _writelines(self, obj):
        """Writes the main body of gwt-like weights files. This code part is
        repeatedly used for many weight file formats. Header lines, however,
        are different from format to format. So, for code reusability, this
        part is separated out from write function by Myunghwa Hwang.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        """

        for id_ in obj.id_order:
            neighbors = list(zip(obj.neighbors[id_], obj.weights[id_], strict=True))
            str_id = "_".join(str(id_).split())
            for neighbor, weight in neighbors:
                neighbor = "_".join(str(neighbor).split())

                self.file.write(f"{str_id} {neighbor} {weight:.6G}\n")
                self.pos += 1

    def write(self, obj):
        """Write a weights object to the opened `GWT` file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A PySAL `W` object.

        Raises
        ------
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(libpysal.examples.get_path('juvenile.gwt'), 'r')
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.gwt')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w')

        Write the weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created ``.gwt`` file.

        >>> wnew =  libpysal.io.open(fname, 'r').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            # transform = obj.transform
            # obj.transform = 'o'

            if hasattr(obj, "_shpName"):
                self.shpName = obj._shpName

            if hasattr(obj, "_varName"):
                self.varName = obj._varName

            header = "%s %i %s %s\n" % ("0", obj.n, self.shpName, self.varName)
            self.file.write(header)
            # obj.transform = transform

            self._writelines(obj)

        else:
            raise TypeError("Expected a PySAL weights object, got: %s." % (type(obj)))

    def close(self):
        self.file.close()
        FileIO.FileIO.close(self)

    @staticmethod
    def __zero_offset(neighbors: dict, weights: dict, original_ids=None) -> dict:
        if not original_ids:
            original_ids = list(neighbors.keys())

        new_weights = {}
        new_ids = {}
        old_ids = {}
        new_neighbors = {}

        for i in original_ids:
            new_i = original_ids.index(i)
            new_ids[new_i] = i
            old_ids[i] = new_i
            neighbors_i = neighbors[i]
            new_neighbors_i = [original_ids.index(j) for j in neighbors_i]
            new_neighbors[new_i] = new_neighbors_i
            new_weights[new_i] = weights[i]

        info = {}
        info["new_ids"] = new_ids
        info["old_ids"] = old_ids
        info["new_neighbors"] = new_neighbors
        info["new_weights"] = new_weights

        return info



# ruff: noqa: N802, N803, N806, N815

from ...weights.util import remap_ids
from ...weights.weights import W
from .. import fileio

__author__ = "Myunghwa Hwang <mhwang4@gmail.com>"
__all__ = ["ArcGISDbfIO"]


class ArcGISDbfIO(fileio.FileIO):
    """Opens, reads, and writes weights file objects in ArcGIS dbf format.

    Spatial weights objects in the ArcGIS ``.dbf`` format are used in ArcGIS Spatial
    Statistics tools. This format is the same as the general ``.dbf`` format,
    but the structure of the weights ``.dbf`` file is fixed unlike other ``.dbf``
    files. This ``.dbf`` format can be used with the "Generate Spatial Weights
    Matrix" tool, but not with the tools under the "Mapping Clusters" category".

    The ArcGIS ``.dbf`` file is assumed to have three or four data columns. When the
    file has four columns, the first column is meaningless and will be ignored in
    PySAL during both file reading and file writing. The next three columns hold
    origin IDs, destinations IDs, and weight values. When the file has three columns,
    it is assumed that only these data columns exist in the stated order. The name
    for the orgin IDs column should be the name of ID variable in the original
    source data table. The names for the destination IDs and weight values columns
    are NID and WEIGHT, respectively. ArcGIS Spatial Statistics tools support only
    unique integer IDs. Therefore, the values for origin and destination ID columns
    should be integer. For the case where the IDs of a weights object are not
    integers, `ArcGISDbfIO` allows users to use internal id values corresponding to
    record numbers, instead of original ids.

    An exemplary structure of an ArcGIS dbf file is as follows:

    ```
    [Line 1]    Field1    RECORD_ID    NID    WEIGHT
    [Line 2]    0         72           76     1
    [Line 3]    0         72           79     1
    [Line 4]    0         72           78     1
    ```

    Unlike the ArcGIS text format, this format does not seem to include self-neighbors.

    References
    ----------

    http://webhelp.esri.com/arcgisdesktop/9.3/index.cfm?TopicName=Convert_Spatial_Weights_Matrix_to_Table_(Spatial_Statistics)

    """

    FORMATS = ["arcgis_dbf"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        self._varName = "Unknown"
        args = args[:2]
        fileio.FileIO.__init__(self, *args, **kwargs)
        self.file = fileio.FileIO(self.dataPath, self.mode)

    def _set_varName(self, val):
        if issubclass(type(val), str):
            self._varName = val

    def _get_varName(self):
        return self._varName

    varName = property(fget=_get_varName, fset=_set_varName)

    def read(self, n=-1):  # noqa: ARG002
        self._complain_ifclosed(self.closed)
        return self._read()

    def seek(self, pos):
        self.file.seek(pos)
        self.pos = self.file.pos

    def _read(self):
        """Reads ArcGIS dbf file

        Returns
        -------
        w : libpysal.weights.W
            A ``libpysal.weights.W`` object.

        Raises
        ------
        StopIteration
            Raised at the EOF.
        ValueError
            Raised when the weights data structure is incorrect.
        TypeError
            Raised when the IDs are not integers.

        Examples
        --------

        Type ``dir(w)`` at the interpreter to see what methods are supported.
        Open an ArcGIS ``.dbf`` file and read it into a PySAL weights object.

        >>> import libpysal
        >>> w = libpysal.io.open(
        ...     libpysal.examples.get_path('arcgis_ohio.dbf'), 'r', 'arcgis_dbf'
        ... ).read()

        Get the number of observations from the header.

        >>> w.n
        88

        Get the mean number of neighbors.

        >>> w.mean_neighbors
        5.25

        Get neighbor distances for a single observation.

        >>> w[1] == dict({2: 1.0, 11: 1.0, 6: 1.0, 7: 1.0})
        True

        """

        if self.pos > 0:
            raise StopIteration

        id_var = self.file.header[1]
        startPos = len(self.file.header)

        if startPos == 3:
            startPos = 0
        elif startPos == 4:
            startPos = 1
        else:
            msg = "Wrong structure, a weights '.dbf' "
            msg += "file requires at least three data columns."
            raise ValueError(msg)

        self.varName = id_var
        id_type = int
        id_spec = self.file.field_spec[startPos]

        if id_spec[0] != "N":
            raise TypeError("The data type for IDs should be integer.")

        self.id_var = id_var

        weights = {}
        neighbors = {}

        for row in self.file:
            i, j, w = tuple(row)[startPos:]
            i = id_type(i)
            j = id_type(j)
            w = float(w)
            if i not in weights:
                weights[i] = []
                neighbors[i] = []
            weights[i].append(w)
            neighbors[i].append(j)
            self.pos = self.file.pos

        w = W(neighbors, weights)

        return w

    def write(self, obj, useIdIndex=False):
        """Write a weights object to the opened ``.dbf`` file.

        Parameters
        ----------
        obj : libpysal.weights.W
            A ``libpysal.weights.W`` object.
        useIdIndex : bool
            Use the `W` IDs and remap (``True``). Default is ``False``.

        Raises
        ------
        TypeError
            Raised when the IDs in input ``obj`` are not integers.
        TypeError
            Raised when the input ``obj`` is not a PySAL `W`.

        Examples
        --------

        >>> import tempfile, libpysal, os
        >>> testfile = libpysal.io.open(
        ...     libpysal.examples.get_path('arcgis_ohio.dbf'), 'r', 'arcgis_dbf'
        ... )
        >>> w = testfile.read()

        Create a temporary file for this example.

        >>> f = tempfile.NamedTemporaryFile(suffix='.dbf')

        Reassign to a new variable.

        >>> fname = f.name

        Close the temporary named file.

        >>> f.close()

        Open the new file in write mode.

        >>> o = libpysal.io.open(fname, 'w', 'arcgis_dbf')

        Write the Weights object into the open file.

        >>> o.write(w)
        >>> o.close()

        Read in the newly created text file.

        >>> wnew =  libpysal.io.open(fname, 'r', 'arcgis_dbf').read()

        Compare values from old to new.

        >>> wnew.pct_nonzero == w.pct_nonzero
        True

        Clean up the temporary file created for this example.

        >>> os.remove(fname)

        """

        self._complain_ifclosed(self.closed)

        if issubclass(type(obj), W):
            self.file.header = [self.varName, "NID", "WEIGHT"]

            id_type = type(obj.id_order[0])
            if id_type is not int and not useIdIndex:
                raise TypeError("ArcGIS DBF weight files support only integer IDs.")
            if useIdIndex:
                id2i = obj.id2i
                obj = remap_ids(obj, id2i)

            id_spec = ("N", len(str(max(obj.id_order))), 0)
            self.file.field_spec = [id_spec, id_spec, ("N", 13, 6)]

            for id_ in obj.id_order:
                neighbors = list(zip(obj.neighbors[id_], obj.weights[id_], strict=True))
                for neighbor, weight in neighbors:
                    self.file.write([id_, neighbor, weight])
                    self.pos = self.file.pos

        else:
            raise TypeError("Expected a PySAL weights object, got: %s." % (type(obj)))

    def flush(self):
        self._complain_ifclosed(self.closed)
        self.file.flush()

    def close(self):
        self.file.close()



# ruff: noqa: N802, N806, N816, N999, SIM115

import datetime
import os
import struct
import time

from ...common import MISSINGVALUE
from .. import tables

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["DBF"]


class DBF(tables.DataTable):
    """PySAL DBF Reader/Writer. This DBF handler implements the PySAL DataTable
    interface and initializes an instance of the PySAL's DBF handler.

    Parameters
    -----------
    dataPath : str
        Path to file, including file name and extension.
    mode : str
        Mode for file interaction; either ``'r'`` or ``'w'``.

    Attributes
    ----------
    header : list
        A list of field names. The header is a python list of strings. Each string
        is a field name and field name must not be longer than 10 characters.
    field_spec : list
        A list describing the data types of each field. It is comprised of a list
        of tuples, each tuple describing a field. The format for the tuples is
        ``('Type', len, precision)``. Valid values for ``'Type'`` are ``'C'``
        for characters, ``'L'`` for bool, ``'D'`` for data, and ``'N'`` or
        ``'F'`` for number.

    Examples
    --------

    >>> import libpysal
    >>> dbf = libpysal.io.open(libpysal.examples.get_path('juvenile.dbf'), 'r')
    >>> dbf.header
    ['ID', 'X', 'Y']

    >>> dbf.field_spec
    [('N', 9, 0), ('N', 9, 0), ('N', 9, 0)]

    """

    FORMATS = ["dbf"]
    MODES = ["r", "w"]

    def __init__(self, *args, **kwargs):
        tables.DataTable.__init__(self, *args, **kwargs)

        if self.mode == "r":
            self.f = f = open(self.dataPath, "rb")
            # from dbf file standards
            numrec, lenheader = struct.unpack("<xxxxLH22x", f.read(32))

            # each field is 32 bytes
            numfields = (lenheader - 33) // 32
            self.n_records = numrec
            self.n_fields = numfields
            self.field_info = [("DeletionFlag", "C", 1, 0)]
            record_size = 1

            # each record is a string
            fmt = "s"

            self._col_index = {}
            idx = 0

            for _ in range(numfields):
                # again, check struct for fmt def.
                name, typ, size, deci = struct.unpack("<11sc4xBB14x", f.read(32))
                # forces to unicode in 2, to str in 3
                name = name.decode()
                typ = typ.decode()
                # same as NULs, \x00
                name = name.replace("\0", "")
                # eliminate NULs from string
                self._col_index[name] = (idx, record_size)
                idx += 1
                # alt: str(size) + 's'
                fmt += "%ds" % size

                record_size += size
                self.field_info.append((name, typ, size, deci))

            terminator = f.read(1).decode()

            assert terminator == "\r"
            self.header_size = self.f.tell()
            self.record_size = record_size
            self.record_fmt = fmt
            self.pos = 0
            self.header = [fInfo[0] for fInfo in self.field_info[1:]]
            field_spec = []

            for _, ftype, flen, fpre in self.field_info[1:]:
                field_spec.append((ftype, flen, fpre))

            self.field_spec = field_spec

            # self.spec = [types[fInfo[0]] for fInfo in self.field_info]

        elif self.mode == "w":
            self.f = open(self.dataPath, "wb")
            self.header = None
            self.field_spec = None
            self.numrec = 0
            self.FIRST_WRITE = True

    def __len__(self) -> int:
        """

        Raises
        ------
        IOError
            Raised when a file is open ``'w'`` mode.

        """

        if self.mode != "r":
            msg = "Invalid operation, cannot read from a file opened in 'w' mode."
            raise OSError(msg)

        return self.n_records

    def seek(self, i):
        self.f.seek(self.header_size + (self.record_size * i))
        self.pos = i

    def _get_col(self, key: str) -> list:
        """Return the column vector.

        Raises
        ------
        AttributeError
            Raised when a field does not exist in the header.

        """

        if key not in self._col_index:
            raise AttributeError("Field: %s does not exist in header." % key)

        prevPos = self.tell()
        idx, offset = self._col_index[key]
        typ, size, deci = self.field_spec[idx]
        gap = self.record_size - size
        f = self.f
        f.seek(self.header_size + offset)
        col = [0] * self.n_records

        for i in range(self.n_records):
            value = f.read(size)
            value = value.decode()
            f.seek(gap, 1)

            if typ == "N":
                value = value.replace("\0", "").lstrip()
                if value == "":
                    value = MISSINGVALUE
                elif deci:
                    try:
                        value = float(value)
                    except ValueError:
                        value = MISSINGVALUE
                else:
                    try:
                        value = int(value)
                    except ValueError:
                        value = MISSINGVALUE
            elif typ == "D":
                try:
                    y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
                    value = datetime.date(y, m, d)
                except ValueError:
                    value = MISSINGVALUE
            elif typ == "L":
                value = (value in "YyTt" and "T") or (value in "NnFf" and "F") or "?"
            elif typ == "F":
                value = value.replace("\0", "").lstrip()
                value = MISSINGVALUE if value == "" else float(value)
            if isinstance(value, str | str):
                value = value.rstrip()
            col[i] = value

        self.seek(prevPos)

        return col

    def read_record(self, i: int) -> list:
        self.seek(i)

        rec = list(struct.unpack(self.record_fmt, self.f.read(self.record_size)))
        rec = [entry.decode() for entry in rec]

        if rec[0] != " ":
            return self.read_record(i + 1)
        result = []

        for (name, typ, _, deci), value in zip(self.field_info, rec, strict=True):
            if name == "DeletionFlag":
                continue
            if typ == "N":
                value = value.replace("\0", "").lstrip()
                if value == "":
                    value = MISSINGVALUE
                elif deci:
                    try:
                        value = float(value)
                    except ValueError:
                        value = MISSINGVALUE
                else:
                    try:
                        value = int(value)
                    except ValueError:
                        value = MISSINGVALUE
            elif typ == "D":
                try:
                    y, m, d = int(value[:4]), int(value[4:6]), int(value[6:8])
                    value = datetime.date(y, m, d)
                except ValueError:
                    # value = datetime.date.min#NULL Date: See issue 114
                    value = MISSINGVALUE
            elif typ == "L":
                value = (value in "YyTt" and "T") or (value in "NnFf" and "F") or "?"
            elif typ == "F":
                value = value.replace("\0", "").lstrip()
                value = MISSINGVALUE if value == "" else float(value)
            if isinstance(value, str | str):
                value = value.rstrip()
            result.append(value)

        return result

    def _read(self) -> list | None:
        """

        Raises
        ------
        IOError
            Raised when a file is open ``'w'`` mode.

        """

        if self.mode != "r":
            msg = "Invalid operation, cannot read from a file opened in 'w' mode."
            raise OSError(msg)

        if self.pos < len(self):
            rec = self.read_record(self.pos)
            self.pos += 1
            return rec
        else:
            return None

    def write(self, obj: list):
        """

        Raises
        ------
        IOError
            Raised when a file is open ``'r'`` mode.
        TypeError
            Raised when a row length and header length are not equivalent.

        """

        self._complain_ifclosed(self.closed)

        if self.mode != "w":
            msg = "Invalid operation, cannot read from a file opened in 'r' mode."
            raise OSError(msg)

        if self.FIRST_WRITE:
            self._firstWrite()

        if len(obj) != len(self.header):
            raise TypeError("Rows must contains %d fields." % len(self.header))

        self.numrec += 1

        # deletion flag
        self.f.write(b" ")

        for (typ, size, deci), value in zip(self.field_spec, obj, strict=True):
            if value is None:
                value = " " * size if typ == "C" else "\x00" * size
            elif typ == "N" or typ == "F":
                # v = str(value).rjust(size, " ")
                # # if len(v) == size:
                # #    value = v
                # # else:
                value = (("%" + "%d.%d" % (size, deci) + "f") % (value))[:size]
            elif typ == "D":
                value = value.strftime("%Y%m%d")
            elif typ == "L":
                value = str(value)[0].upper()
            else:
                value = str(value)[:size].ljust(size, " ")
            try:
                assert len(value) == size
            except:
                print(value, len(value), size)
                raise
            self.f.write(value.encode())
            self.pos += 1

    def flush(self):
        self._complain_ifclosed(self.closed)
        self._writeHeader()
        self.f.flush()

    def close(self):
        if self.mode == "w":
            self.flush()
            # End of file
            self.f.write(b"\x1A")
        self.f.close()

        tables.DataTable.close(self)

    def _firstWrite(self):
        """

        Raises
        ------
        IOError
            Raised when there is no specified header.
        IOError
            Raised when there is no field specification.

        """

        if not self.header:
            raise OSError("No header, DBF files require a header.")
        if not self.field_spec:
            raise OSError("No field_spec, DBF files require a specification.")

        self._writeHeader()

        self.FIRST_WRITE = False

    def _writeHeader(self):
        """Modified from:
        http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/362715
        """

        POS = self.f.tell()
        self.f.seek(0)
        ver = 3
        now = datetime.datetime.utcfromtimestamp(
            int(os.environ.get("SOURCE_DATE_EPOCH", time.time())),
        )

        yr, mon, day = now.year - 1900, now.month, now.day
        numrec = self.numrec
        numfields = len(self.header)
        lenheader = numfields * 32 + 33
        lenrecord = sum(field[1] for field in self.field_spec) + 1

        hdr = struct.pack(
            "<BBBBLHH20x", ver, yr, mon, day, numrec, lenheader, lenrecord
        )
        self.f.write(hdr)

        # field specs
        for name, (typ, size, deci) in zip(self.header, self.field_spec, strict=True):
            typ = typ.encode()
            name = name.ljust(11, "\x00")
            name = name.encode()
            fld = struct.pack("<11sc4xBB14x", name, typ, size, deci)
            self.f.write(fld)

        # terminator
        term = b"\r"
        self.f.write(term)
        if self.f.tell() != POS and not self.FIRST_WRITE:
            self.f.seek(POS)


if __name__ == "__main__":
    import libpysal

    file_name = libpysal.examples.get_path("10740.dbf")
    f = libpysal.open(file_name, "r")
    newDB = libpysal.open("copy.dbf", "w")

    newDB.header = f.header
    newDB.field_spec = f.field_spec
    print(f.header)
    for row in f:
        print(row)
        newDB.write(row)
    newDB.close()
    copy = libpysal.open("copy.dbf", "r")

    f.seek(0)
    print("HEADER: ", copy.header == f.header)
    print("SPEC: ", copy.field_spec == f.field_spec)
    print("DATA: ", list(copy) == list(f))



print("import: 'libpysal'")
import libpysal

print("import: 'libpysal.cg'")
import libpysal.cg

print("import: 'libpysal.cg.ops'")
import libpysal.cg.ops

print("import: 'libpysal.examples'")
import libpysal.examples

print("import: 'libpysal.io'")
import libpysal.io

print("import: 'libpysal.io.geotable'")
import libpysal.io.geotable

print("import: 'libpysal.io.iohandlers'")
import libpysal.io.iohandlers

print("import: 'libpysal.io.tests'")
import libpysal.io.tests

print("import: 'libpysal.io.util'")
import libpysal.io.util

print("import: 'libpysal.weights'")
import libpysal.weights




import functools as _f
from warnings import warn

from ...common import requires as _requires
from ...io.geotable.utils import to_df, to_gdf

try:
    import pandas as _pd

    @_requires("pandas")
    @_f.wraps(_pd.merge)
    def join(*args, **kwargs):
        return _pd.merge(*args, **kwargs)

except ImportError:
    pass


@_requires("geopandas")
def spatial_join(
    df1, df2, left_geom_col="geometry", right_geom_col="geometry", **kwargs
):
    """Perform a spatial join between two ``pandas.DataFrames``
    datasets by calling out to ``geopandas``.

    Parameters
    ----------
    df1 : pandas.DataFrame
        The first dataset. It must have a 'MultiPolygon'
        or 'Polygon' geometry column.
    df2 : pandas.DataFrame
        The second dataset. It must have a 'MultiPolygon'
        or 'Polygon' geometry column.
    left_geom_col : str
        The left (``df1``) dataset's geometry column name.
        Default is ``'geometry'``.
    right_geom_col : str
        The right (``df2``) dataset's geometry column name.
        Default is ``'geometry'``.
    **kwargs : dict
        Optional keyword arguments passed in ``geopandas.tools.sjoin``.
        These may include (1) ``'how'`` (the method of spatial join),
        with valid values including ``'left'`` (use keys from ``df1``
        and retain only the ``df1`` geometry column), ``'right'`` (use
        keys from ``df2`` and retain only the ``df2`` geometry column,
        and ``'inner'`` (use the intersection of keys from both ``df1``
        & ``df2`` and retain only the `df2`` geometry column.;
        (2) ``'op'`` (defaults to ``'intersects'``), with other valid
        values including ``'contains'`` and ``'within'``. See the
        `Shapely docs <http://toblerity.org/shapely/manual.html#binary-predicates>`_
        for more information.; (3) ``'lsuffix'`` (defaults to ``left'``),
        the suffix to apply to overlapping column names from ``df1``.;
        and (4) ``'rsuffix'`` defaults to ``right'``),
        the suffix to apply to overlapping column names from ``df2``.

    Returns
    -------
    df : pandas.DataFrame
        A pandas.DataFrame with a new set of polygons
        and attributes resulting from the overlay.
    """

    import geopandas as gpd

    gdf1 = to_gdf(df1, geom_col=left_geom_col)
    gdf2 = to_gdf(df2, geom_col=right_geom_col)
    out = gpd.tools.sjoin(gdf1, gdf2, **kwargs)

    df = to_df(out)

    return df


@_requires("geopandas")
def spatial_overlay(
    df1, df2, how, left_geom_col="geometry", right_geom_col="geometry", **kwargs
):
    """Perform a spatial overlay between two polygonal
    datasets by calling out to ``geopandas``. It currently
    only supports data (``pandas.DataFrames``) with polygons
    and implements several methods that are all effectively
    subsets of the union.

    Parameters
    ----------
    df1 : pandas.DataFrame
        The first dataset. It must have a 'MultiPolygon'
        or 'Polygon' geometry column.
    df2 : pandas.DataFrame
        The second dataset. It must have a 'MultiPolygon'
        or 'Polygon' geometry column.
    how : str
        The method of spatial overlay. Options are inculde
        ``'intersection'``, ``'union', ``'identity'``,
        ``'symmetric_difference'`` or ``'difference'``.
    left_geom_col : str
        The left (``df1``) dataset's geometry column name.
        Default is ``'geometry'``.
    right_geom_col : str
        The right (``df2``) dataset's geometry column name.
        Default is ``'geometry'``.
    **kwargs : dict
        Optional keyword arguments passed in ``geopandas.tools.overlay``.

    Returns
    -------
    df : pandas.DataFrame
        A pandas.DataFrame with a new set of polygons
        and attributes resulting from the overlay.
    """

    import geopandas as gpd

    gdf1 = to_gdf(df1, geom_col=left_geom_col)
    gdf2 = to_gdf(df2, geom_col=right_geom_col)
    out = gpd.tools.overlay(gdf1, gdf2, how, **kwargs)

    df = to_df(out)

    return df


@_requires("shapely")
def dissolve(df, by="", **groupby_kws):
    from ._shapely import cascaded_union as union

    return union(df, by=by, **groupby_kws)


def clip(return_exterior=False):  # noqa: ARG001
    # return modified entries of the df that are within an envelope
    # provide an option to null out the geometries instead of not returning
    raise NotImplementedError


def erase(return_interior=True):  # noqa: ARG001
    # return modified entries of the df that are outside of an envelope
    # provide an option to null out the geometries instead of not returning
    raise NotImplementedError


@_requires("shapely")
def union(df, **kws):
    if "by" in kws:
        warn(
            "When a 'by' argument is provided, you should be using 'dissolve'.",
            stacklevel=2,
        )
        return dissolve(df, **kws)
    from ._shapely import cascaded_union as union

    return union(df)


@_requires("shapely")
def intersection(df, **kws):
    from ._shapely import cascaded_intersection as intersection

    return intersection(df, **kws)


def symmetric_difference():
    raise NotImplementedError


def difference():
    raise NotImplementedError



"""
Helper functions for computational geometry in PySAL.

"""

__author__ = "Sergio J. Rey, Xinyue Ye, Charles Schmidt, Andrew Winslow"
__credits__ = "Copyright (c) 2005-2009 Sergio J. Rey"

# ruff: noqa: F403, F405, N803


import copy
import math
import random
from itertools import islice

import numpy as np
import scipy.spatial

from .shapes import *

EPSILON_SCALER = 3


__all__ = [
    "bbcommon",
    "get_bounding_box",
    "get_angle_between",
    "is_collinear",
    "get_segments_intersect",
    "get_segment_point_intersect",
    "get_polygon_point_intersect",
    "get_rectangle_point_intersect",
    "get_ray_segment_intersect",
    "get_rectangle_rectangle_intersection",
    "get_polygon_point_dist",
    "get_points_dist",
    "get_segment_point_dist",
    "get_point_at_angle_and_dist",
    "convex_hull",
    "is_clockwise",
    "point_touches_rectangle",
    "get_shared_segments",
    "distance_matrix",
]


def bbcommon(bb, bbother):
    """Old Stars method for bounding box overlap testing.
    Also defined in ``pysal.weights._cont_binning``.

    Parameters
    ----------
    bb : list
        A bounding box.
    bbother : list
        The bounding box to test against.

    Returns
    -------
    chflag : int
        ``1`` if ``bb`` overlaps ``bbother``, otherwise ``0``.

    Examples
    --------
    >>> b0 = [0, 0, 10, 10]
    >>> b1 = [10, 0, 20, 10]
    >>> bbcommon(b0, b1)
    1
    """

    chflag = 0

    if (not ((bbother[2] < bb[0]) or (bbother[0] > bb[2]))) and (
        not ((bbother[3] < bb[1]) or (bbother[1] > bb[3]))
    ):
        chflag = 1

    return chflag


def get_bounding_box(items):
    """Find bounding box for a list of geometries.

    Parameters
    ----------
    items : list
        PySAL shapes.

    Returns
    -------
    rect = libpysal.cg.Rectangle
        The bounding box for a list of geometries.

    Examples
    --------
    >>> bb = get_bounding_box([Point((-1, 5)), Rectangle(0, 6, 11, 12)])
    >>> bb.left
    -1.0

    >>> bb.lower
    5.0

    >>> bb.right
    11.0

    >>> bb.upper
    12.0
    """

    def left(o):
        # Polygon, Ellipse
        if hasattr(o, "bounding_box"):
            return o.bounding_box.left
        # Rectangle
        elif hasattr(o, "left"):
            return o.left
        # Point
        else:
            return o[0]

    def right(o):
        # Polygon, Ellipse
        if hasattr(o, "bounding_box"):
            return o.bounding_box.right
        # Rectangle
        elif hasattr(o, "right"):
            return o.right
        # Point
        else:
            return o[0]

    def lower(o):
        # Polygon, Ellipse
        if hasattr(o, "bounding_box"):
            return o.bounding_box.lower
        # Rectangle
        elif hasattr(o, "lower"):
            return o.lower
        # Point
        else:
            return o[1]

    def upper(o):
        # Polygon, Ellipse
        if hasattr(o, "bounding_box"):
            return o.bounding_box.upper
        # Rectangle
        elif hasattr(o, "upper"):
            return o.upper
        # Point
        else:
            return o[1]

    rect = Rectangle(
        min(list(map(left, items))),
        min(list(map(lower, items))),
        max(list(map(right, items))),
        max(list(map(upper, items))),
    )

    return rect


def get_angle_between(ray1, ray2):
    """Returns the angle formed between a pair of rays which share an origin.

    Parameters
    ----------
    ray1 : libpysal.cg.Ray
        A ray forming the beginning of the angle measured.
    ray2 : libpysal.cg.Ray
        A ray forming the end of the angle measured.

    Returns
    -------
    angle : float
        The angle between ``ray1`` and ``ray2``.

    Raises
    ------
    ValueError
        Raised when rays do not have the same origin.

    Examples
    --------
    >>> get_angle_between(
    ...     Ray(Point((0, 0)), Point((1, 0))),
    ...     Ray(Point((0, 0)), Point((1, 0)))
    ... )
    0.0
    """

    if ray1.o != ray2.o:
        raise ValueError("Rays must have the same origin.")

    vec1 = (ray1.p[0] - ray1.o[0], ray1.p[1] - ray1.o[1])
    vec2 = (ray2.p[0] - ray2.o[0], ray2.p[1] - ray2.o[1])

    rot_theta = -math.atan2(vec1[1], vec1[0])
    rot_matrix = [
        [math.cos(rot_theta), -math.sin(rot_theta)],
        [math.sin(rot_theta), math.cos(rot_theta)],
    ]

    rot_vec2 = (
        rot_matrix[0][0] * vec2[0] + rot_matrix[0][1] * vec2[1],
        rot_matrix[1][0] * vec2[0] + rot_matrix[1][1] * vec2[1],
    )

    angle = math.atan2(rot_vec2[1], rot_vec2[0])

    return angle


def is_collinear(p1, p2, p3):
    """Returns whether a triplet of points is collinear.

    Parameters
    ----------
    p1 : libpysal.cg.Point
        A point.
    p2 : libpysal.cg.Point
        A point.
    p3 : libpysal.cg.Point
        A point.

    Returns
    -------
    collinear : bool
        ``True`` if ``{p1, p2, p3}`` are collinear, otherwise ``False``.

    Examples
    --------
    >>> is_collinear(Point((0, 0)), Point((1, 1)), Point((5, 5)))
    True

    >>> is_collinear(Point((0, 0)), Point((1, 1)), Point((5, 0)))
    False
    """

    eps = np.finfo(type(p1[0])).eps

    slope_diff = abs(
        (p2[0] - p1[0]) * (p3[1] - p1[1]) - (p2[1] - p1[1]) * (p3[0] - p1[0])
    )

    very_small_dist = EPSILON_SCALER * eps

    collinear = slope_diff < very_small_dist

    return collinear


def get_segments_intersect(seg1, seg2):
    """Returns the intersection of two segments if one exists.

    Parameters
    ----------
    seg1 : libpysal.cg.LineSegment
        A segment to check for an intersection.
    seg2 : libpysal.cg.LineSegment
        The segment to check against ``seg1`` for an intersection.

    Returns
    -------
    intersection : {libpysal.cg.Point, libpysal.cg.LineSegment, None}
        The intersecting point or line between ``seg1`` and
        ``seg2`` if an intersection exists or ``None`` if
        ``seg1`` and ``seg2`` do not intersect.

    Examples
    --------
    >>> seg1 = LineSegment(Point((0, 0)), Point((0, 10)))
    >>> seg2 = LineSegment(Point((-5, 5)), Point((5, 5)))
    >>> i = get_segments_intersect(seg1, seg2)
    >>> isinstance(i, Point)
    True

    >>> str(i)
    '(0.0, 5.0)'

    >>> seg3 = LineSegment(Point((100, 100)), Point((100, 101)))
    >>> i = get_segments_intersect(seg2, seg3)
    """

    p1 = seg1.p1
    p2 = seg1.p2
    p3 = seg2.p1
    p4 = seg2.p2
    a = p2[0] - p1[0]
    b = p3[0] - p4[0]
    c = p2[1] - p1[1]
    d = p3[1] - p4[1]
    det = float(a * d - b * c)

    intersection = None

    if det == 0:
        if seg1 == seg2:
            intersection = LineSegment(seg1.p1, seg1.p2)
        else:
            a = get_segment_point_intersect(seg2, seg1.p1)
            b = get_segment_point_intersect(seg2, seg1.p2)
            c = get_segment_point_intersect(seg1, seg2.p1)
            d = get_segment_point_intersect(seg1, seg2.p2)
            if a and b:  # seg1 in seg2
                intersection = LineSegment(seg1.p1, seg1.p2)
            if c and d:  # seg2 in seg1
                intersection = LineSegment(seg2.p1, seg2.p2)
            if (a or b) and (c or d):
                p1 = a if a else b
                p2 = c if c else d
                intersection = LineSegment(p1, p2)
    else:
        a_inv = d / det
        b_inv = -b / det
        c_inv = -c / det
        d_inv = a / det
        m = p3[0] - p1[0]
        n = p3[1] - p1[1]
        x = a_inv * m + b_inv * n
        y = c_inv * m + d_inv * n
        intersect_exists = 0 <= x <= 1 and 0 <= y <= 1

        if intersect_exists:
            intersection = Point(
                (p1[0] + x * (p2[0] - p1[0]), p1[1] + x * (p2[1] - p1[1]))
            )

    return intersection


def get_segment_point_intersect(seg, pt):
    """Returns the intersection of a segment and point.

    Parameters
    ----------
    seg : libpysal.cg.LineSegment
        A segment to check for an intersection.
    pt : libpysal.cg.Point
        A point to check ``seg`` for an intersection.

    Returns
    -------
    pt : {libpysal.cg.Point, None}
        The intersection of a ``seg`` and ``pt`` if one exists, otherwise ``None``.

    Examples
    --------
    >>> seg = LineSegment(Point((0, 0)), Point((0, 10)))
    >>> pt = Point((0, 5))
    >>> i = get_segment_point_intersect(seg, pt)
    >>> str(i)
    '(0.0, 5.0)'

    >>> pt2 = Point((5, 5))
    >>> get_segment_point_intersect(seg, pt2)
    """

    eps = np.finfo(type(pt[0])).eps

    if is_collinear(pt, seg.p1, seg.p2):
        if get_segment_point_dist(seg, pt)[0] < EPSILON_SCALER * eps:
            pass
        else:
            pt = None
    else:
        vec1 = (pt[0] - seg.p1[0], pt[1] - seg.p1[1])
        vec2 = (seg.p2[0] - seg.p1[0], seg.p2[1] - seg.p1[1])

        if abs(vec1[0] * vec2[1] - vec1[1] * vec2[0]) < eps:
            pass
        else:
            pt = None

    return pt


def get_polygon_point_intersect(poly, pt):
    """Returns the intersection of a polygon and point.

    Parameters
    ----------
    poly : libpysal.cg.Polygon
        A polygon to check for an intersection.
    pt : libpysal.cg.Point
        A point to check ``poly`` for an intersection.

    Returns
    -------
    ret : {libpysal.cg.Point, None}
        The intersection of a ``poly`` and ``pt`` if one exists, otherwise ``None``.

    Examples
    --------
    >>> poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
    >>> pt = Point((0.5, 0.5))
    >>> i = get_polygon_point_intersect(poly, pt)
    >>> str(i)
    '(0.5, 0.5)'

    >>> pt2 = Point((2, 2))
    >>> get_polygon_point_intersect(poly, pt2)
    """

    def pt_lies_on_part_boundary(p, vx):
        vx_range = range(-1, len(vx) - 1)
        seg = lambda i: LineSegment(vx[i], vx[i + 1])  # noqa: E731
        return [i for i in vx_range if get_segment_point_dist(seg(i), p)[0] == 0] != []

    ret = None

    # Weed out points that aren't even close
    if get_rectangle_point_intersect(poly.bounding_box, pt) is None:
        pass
    else:
        if [
            vxs for vxs in poly._vertices if pt_lies_on_part_boundary(pt, vxs)
        ] != [] or [vxs for vxs in poly._vertices if _point_in_vertices(pt, vxs)] != []:
            ret = pt
        if poly._holes != [[]]:
            if [vxs for vxs in poly.holes if pt_lies_on_part_boundary(pt, vxs)] != []:
                # pt lies on boundary of hole.
                pass
            if [vxs for vxs in poly.holes if _point_in_vertices(pt, vxs)] != []:
                # pt lines inside a hole.
                ret = None
            # raise NotImplementedError,
            # 'Cannot compute containment for polygon with holes'

    return ret


def get_rectangle_point_intersect(rect, pt):
    """Returns the intersection of a rectangle and point.

    Parameters
    ----------
    rect : libpysal.cg.Rectangle
        A rectangle to check for an intersection.
    pt : libpysal.cg.Point
        A point to check ``rect`` for an intersection.

    Returns
    -------
    pt : {libpysal.cg.Point, None}
        The intersection of a ``rect`` and ``pt`` if one exists, otherwise ``None``.

    Examples
    --------

    >>> rect = Rectangle(0, 0, 5, 5)
    >>> pt = Point((1, 1))
    >>> i = get_rectangle_point_intersect(rect, pt)
    >>> str(i)
    '(1.0, 1.0)'

    >>> pt2 = Point((10, 10))
    >>> get_rectangle_point_intersect(rect, pt2)
    """

    if rect.left <= pt[0] <= rect.right and rect.lower <= pt[1] <= rect.upper:
        pass
    else:
        pt = None

    return pt


def get_ray_segment_intersect(ray, seg):
    """Returns the intersection of a ray and line segment.

    Parameters
    ----------
    ray : libpysal.cg.Ray
        A ray to check for an intersection.
    seg : libpysal.cg.LineSegment
        A segment to check for an intersection against ``ray``.

    Returns
    -------
    intersection : {libpysal.cg.Point, libpysal.cg.LineSegment, None}
        The intersecting point or line between ``ray`` and
        ``seg`` if an intersection exists or ``None`` if
        ``ray`` and ``seg`` do not intersect.

    See Also
    --------
    libpysal.cg.get_segments_intersect

    Examples
    --------
    >>> ray = Ray(Point((0, 0)), Point((0, 1)))
    >>> seg = LineSegment(Point((-1, 10)), Point((1, 10)))
    >>> i = get_ray_segment_intersect(ray, seg)
    >>> isinstance(i, Point)
    True

    >>> str(i)
    '(0.0, 10.0)'

    >>> seg2 = LineSegment(Point((10, 10)), Point((10, 11)))
    >>> get_ray_segment_intersect(ray, seg2)
    """

    # Upper bound on origin to segment dist (+1)
    d = (
        max(
            math.hypot(seg.p1[0] - ray.o[0], seg.p1[1] - ray.o[1]),
            math.hypot(seg.p2[0] - ray.o[0], seg.p2[1] - ray.o[1]),
        )
        + 1
    )
    ratio = d / math.hypot(ray.o[0] - ray.p[0], ray.o[1] - ray.p[1])
    ray_seg = LineSegment(
        ray.o,
        Point(
            (
                ray.o[0] + ratio * (ray.p[0] - ray.o[0]),
                ray.o[1] + ratio * (ray.p[1] - ray.o[1]),
            )
        ),
    )

    intersection = get_segments_intersect(seg, ray_seg)

    return intersection


def get_rectangle_rectangle_intersection(r0, r1, checkOverlap=True):
    """Returns the intersection between two rectangles.

    Parameters
    ----------
    r0 : libpysal.cg.Rectangle
        A rectangle to check for an intersection.
    r1 : libpysal.cg.Rectangle
        A rectangle to check for an intersection against ``r0``.
    checkOverlap : bool
        Call ``bbcommon(r0, r1)`` prior to complex geometry
        checking. Default is ``True``. Prior to setting as
        ``False`` see the Notes section.

    Returns
    -------
    intersection : {libpysal.cg.Point, libpysal.cg.LineSegment, libpysal.cg.Rectangle, None}
        The intersecting point, line, or rectangle between
        ``r0`` and ``r1`` if an intersection exists or ``None``
        if ``r0`` and ``r1`` do not intersect.

    Notes
    -----
    The algorithm assumes the rectangles overlap. The keyword
    ``checkOverlap=False`` should be used with extreme caution.

    Examples
    --------
    >>> r0 = Rectangle(0,4,6,9)
    >>> r1 = Rectangle(4,0,9,7)
    >>> ri = get_rectangle_rectangle_intersection(r0,r1)
    >>> ri[:]
    [4.0, 4.0, 6.0, 7.0]

    >>> r0 = Rectangle(0,0,4,4)
    >>> r1 = Rectangle(2,1,6,3)
    >>> ri = get_rectangle_rectangle_intersection(r0,r1)
    >>> ri[:]
    [2.0, 1.0, 4.0, 3.0]

    >>> r0 = Rectangle(0,0,4,4)
    >>> r1 = Rectangle(2,1,3,2)
    >>> ri = get_rectangle_rectangle_intersection(r0,r1)
    >>> ri[:] == r1[:]
    True
    """  # noqa: E501

    intersection = None
    common_bb = True

    if checkOverlap and not bbcommon(r0, r1):
        # raise ValueError, "Rectangles do not intersect"
        common_bb = False

    if common_bb:
        left = max(r0.left, r1.left)
        lower = max(r0.lower, r1.lower)
        right = min(r0.right, r1.right)
        upper = min(r0.upper, r1.upper)

        if upper == lower and left == right:
            intersection = Point((left, lower))
        elif upper == lower:
            intersection = LineSegment(Point((left, lower)), Point((right, lower)))
        elif left == right:
            intersection = LineSegment(Point((left, lower)), Point((left, upper)))
        else:
            intersection = Rectangle(left, lower, right, upper)

    return intersection


def get_polygon_point_dist(poly, pt):
    """Returns the distance between a polygon and point.

    Parameters
    ----------
    poly : libpysal.cg.Polygon
        A polygon to compute distance from.

    pt : libpysal.cg.Point
        a point to compute distance from

    Returns
    -------
    dist : float
        The distance between ``poly`` and ``point``.

    Examples
    --------
    >>> poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
    >>> pt = Point((2, 0.5))
    >>> get_polygon_point_dist(poly, pt)
    1.0

    >>> pt2 = Point((0.5, 0.5))
    >>> get_polygon_point_dist(poly, pt2)
    0.0
    """

    if get_polygon_point_intersect(poly, pt) is not None:
        dist = 0.0
    else:
        part_prox = []
        for vertices in poly._vertices:
            vx_range = range(-1, len(vertices) - 1)
            seg = lambda i: LineSegment(  # noqa: E731
                vertices[i],  # noqa: B023
                vertices[i + 1],  # noqa: B023
            )
            _min_dist = min([get_segment_point_dist(seg(i), pt)[0] for i in vx_range])
            part_prox.append(_min_dist)
        dist = min(part_prox)

    return dist


def get_points_dist(pt1, pt2):
    """Returns the distance between a pair of points.

    Parameters
    ----------
    pt1 : libpysal.cg.Point
        A point.

    pt2 : libpysal.cg.Point
        The other point.

    Returns
    -------
    dist : float
        The distance between ``pt1`` and ``pt2``.

    Examples
    --------
    >>> get_points_dist(Point((4, 4)), Point((4, 8)))
    4.0

    >>> get_points_dist(Point((0, 0)), Point((0, 0)))
    0.0
    """

    dist = math.hypot(pt1[0] - pt2[0], pt1[1] - pt2[1])

    return dist


def get_segment_point_dist(seg, pt):
    """Returns (1) the distance between a line segment and point
    and (2) the distance along the segment to the closest location on the
    segment from the point as a ratio of the length of the segment.

    Parameters
    ----------
    seg : libpysal.cg.LineSegment
        A line segment to compute distance from.
    pt : libpysal.cg.Point
        A point to compute distance from.

    Returns
    -------
    dist : float
        The distance between ``seg`` and ``pt``.
    ratio : float
        The distance along ``seg`` to the closest location on
        ``seg`` from ``pt`` as a ratio of the length of ``seg``.

    Examples
    --------
    >>> seg = LineSegment(Point((0, 0)), Point((10, 0)))
    >>> pt = Point((5, 5))
    >>> get_segment_point_dist(seg, pt)
    (5.0, 0.5)

    >>> pt2 = Point((0, 0))
    >>> get_segment_point_dist(seg, pt2)
    (0.0, 0.0)
    """

    src_p = seg.p1
    dest_p = seg.p2

    # Shift line to go through origin
    points_0 = pt[0] - src_p[0]
    points_1 = pt[1] - src_p[1]
    points_2 = 0
    points_3 = 0
    points_4 = dest_p[0] - src_p[0]
    points_5 = dest_p[1] - src_p[1]

    segment_length = get_points_dist(src_p, dest_p)

    # Meh, robustness...
    # maybe should incorporate this into a more general approach later
    if segment_length == 0:
        dist, ratio = get_points_dist(pt, src_p), 0

    else:
        u_x = points_4 / segment_length
        u_y = points_5 / segment_length

        inter_x = u_x * u_x * points_0 + u_x * u_y * points_1
        inter_y = u_x * u_y * points_0 + u_y * u_y * points_1

        src_proj_dist = get_points_dist((0, 0), (inter_x, inter_y))
        dest_proj_dist = get_points_dist((inter_x, inter_y), (points_4, points_5))

        if src_proj_dist > segment_length or dest_proj_dist > segment_length:
            src_pt_dist = get_points_dist((points_2, points_3), (points_0, points_1))
            dest_pt_dist = get_points_dist((points_4, points_5), (points_0, points_1))

            if src_pt_dist < dest_pt_dist:
                dist, ratio = src_pt_dist, 0
            else:
                dist, ratio = dest_pt_dist, 1
        else:
            dist = get_points_dist((inter_x, inter_y), (points_0, points_1))
            ratio = src_proj_dist / segment_length

    return dist, ratio


def get_point_at_angle_and_dist(ray, angle, dist):
    """Returns the point at a distance and angle relative to the origin of a ray.

    Parameters
    ----------
    ray : libpysal.cg.Ray
        The ray to which ``angle`` and ``dist`` are relative.
    angle : float
        The angle relative to ``ray`` at which ``point`` is located.
    dist : float
        The distance from the origin of ``ray`` at which ``point`` is located.

    Returns
    -------
    point : libpysal.cg.Point
        The point at ``dist`` and ``angle`` relative to the origin of ``ray``.

    Examples
    --------
    >>> ray = Ray(Point((0, 0)), Point((1, 0)))
    >>> pt = get_point_at_angle_and_dist(ray, math.pi, 1.0)
    >>> isinstance(pt, Point)
    True

    >>> round(pt[0], 8)
    -1.0

    >>> round(pt[1], 8)
    0.0
    """

    v = (ray.p[0] - ray.o[0], ray.p[1] - ray.o[1])
    cur_angle = math.atan2(v[1], v[0])
    dest_angle = cur_angle + angle

    point = Point(
        (ray.o[0] + dist * math.cos(dest_angle), ray.o[1] + dist * math.sin(dest_angle))
    )

    return point


def convex_hull(points):
    """Returns the convex hull of a set of points.

    Parameters
    ----------
    points : list
        A list of points for computing the convex hull.

    Returns
    -------
    stack : list
        A list of points representing the convex hull.

    Examples
    --------
    >>> points = [Point((0, 0)), Point((4, 4)), Point((4, 0)), Point((3, 1))]
    >>> convex_hull(points)
    [(0.0, 0.0), (4.0, 0.0), (4.0, 4.0)]
    """

    def right_turn(p1, p2, p3) -> bool:
        """Returns if ``p1`` -> ``p2`` -> ``p3`` forms a 'right turn'."""
        vec1 = (p2[0] - p1[0], p2[1] - p1[1])
        vec2 = (p3[0] - p2[0], p3[1] - p2[1])
        _rt = vec2[0] * vec1[1] - vec2[1] * vec1[0] >= 0
        return _rt

    points = copy.copy(points)
    lowest = min(points, key=lambda p: (p[1], p[0]))

    points.remove(lowest)
    points.sort(key=lambda p: math.atan2(p[1] - lowest[1], p[0] - lowest[0]))

    stack = [lowest]

    for p in points:
        stack.append(p)
        while len(stack) > 3 and right_turn(stack[-3], stack[-2], stack[-1]):
            stack.pop(-2)

    return stack


def is_clockwise(vertices):
    """Returns whether a list of points describing
    a polygon are clockwise or counterclockwise.

    Parameters
    ----------
    vertices : list
        A list of points that form a single ring.

    Returns
    -------
    clockwise : bool
        ``True`` if ``vertices`` are clockwise, otherwise ``False``.

    See Also
    --------
    libpysal.cg.ccw

    Examples
    --------
    >>> is_clockwise([Point((0, 0)), Point((10, 0)), Point((0, 10))])
    False

    >>> is_clockwise([Point((0, 0)), Point((0, 10)), Point((10, 0))])
    True

    >>> v = [
    ...     (-106.57798, 35.174143999999998),
    ...     (-106.583412, 35.174141999999996),
    ...     (-106.58417999999999, 35.174143000000001),
    ...     (-106.58377999999999, 35.175542999999998),
    ...     (-106.58287999999999, 35.180543),
    ...     (-106.58263099999999, 35.181455),
    ...     (-106.58257999999999, 35.181643000000001),
    ...     (-106.58198299999999, 35.184615000000001),
    ...     (-106.58148, 35.187242999999995),
    ...     (-106.58127999999999, 35.188243),
    ...     (-106.58138, 35.188243),
    ...     (-106.58108, 35.189442999999997),
    ...     (-106.58104, 35.189644000000001),
    ...     (-106.58028, 35.193442999999995),
    ...     (-106.580029, 35.194541000000001),
    ...     (-106.57974399999999, 35.195785999999998),
    ...     (-106.579475, 35.196961999999999),
    ...     (-106.57922699999999, 35.198042999999998),
    ...     (-106.578397, 35.201665999999996),
    ...     (-106.57827999999999, 35.201642999999997),
    ...     (-106.57737999999999, 35.201642999999997),
    ...     (-106.57697999999999, 35.201543000000001),
    ...     (-106.56436599999999, 35.200311999999997),
    ...     (-106.56058, 35.199942999999998),
    ...     (-106.56048, 35.197342999999996),
    ...     (-106.56048, 35.195842999999996),
    ...     (-106.56048, 35.194342999999996),
    ...     (-106.56048, 35.193142999999999),
    ...     (-106.56048, 35.191873999999999),
    ...     (-106.56048, 35.191742999999995),
    ...     (-106.56048, 35.190242999999995),
    ...     (-106.56037999999999, 35.188642999999999),
    ...     (-106.56037999999999, 35.187242999999995),
    ...     (-106.56037999999999, 35.186842999999996),
    ...     (-106.56037999999999, 35.186552999999996),
    ...     (-106.56037999999999, 35.185842999999998),
    ...     (-106.56037999999999, 35.184443000000002),
    ...     (-106.56037999999999, 35.182943000000002),
    ...     (-106.56037999999999, 35.181342999999998),
    ...     (-106.56037999999999, 35.180433000000001),
    ...     (-106.56037999999999, 35.179943000000002),
    ...     (-106.56037999999999, 35.178542999999998),
    ...     (-106.56037999999999, 35.177790999999999),
    ...     (-106.56037999999999, 35.177143999999998),
    ...     (-106.56037999999999, 35.175643999999998),
    ...     (-106.56037999999999, 35.174444000000001),
    ...     (-106.56037999999999, 35.174043999999995),
    ...     (-106.560526, 35.174043999999995),
    ...     (-106.56478, 35.174043999999995),
    ...     (-106.56627999999999, 35.174143999999998),
    ...     (-106.566541, 35.174144999999996),
    ...     (-106.569023, 35.174157000000001),
    ...     (-106.56917199999999, 35.174157999999998),
    ...     (-106.56938, 35.174143999999998),
    ...     (-106.57061499999999, 35.174143999999998),
    ...     (-106.57097999999999, 35.174143999999998),
    ...     (-106.57679999999999, 35.174143999999998),
    ...     (-106.57798, 35.174143999999998)
    ... ]
    >>> is_clockwise(v)
    True
    """

    clockwise = True

    if not len(vertices) < 3:
        area = 0.0
        ax, ay = vertices[0]
        for bx, by in vertices[1:]:
            area += ax * by - ay * bx
            ax, ay = bx, by
        bx, by = vertices[0]
        area += ax * by - ay * bx

        clockwise = area < 0.0

    return clockwise


def ccw(vertices):
    """Returns whether a list of points is counterclockwise.

    Parameters
    ----------
    vertices : list
        A list of points that form a single ring.

    Returns
    -------
    counter_clockwise : bool
        ``True`` if ``vertices`` are counter clockwise, otherwise ``False``.

    See Also
    --------
    libpysal.cg.is_clockwise

    Examples
    --------
    >>> ccw([Point((0, 0)), Point((10, 0)), Point((0, 10))])
    True

    >>> ccw([Point((0, 0)), Point((0, 10)), Point((10, 0))])
    False
    """

    counter_clockwise = True

    if is_clockwise(vertices):
        counter_clockwise = False

    return counter_clockwise


def seg_intersect(a, b, c, d):
    """Tests if two segments (a,b) and (c,d) intersect.

    Parameters
    ----------
    a : libpysal.cg.Point
        The first vertex for the first segment.
    b : libpysal.cg.Point
        The second vertex for the first segment.
    c : libpysal.cg.Point
        The first vertex for the second segment.
    d : libpysal.cg.Point
        The second vertex for the second segment.

    Returns
    -------
    segments_intersect : bool
        ``True`` if segments ``(a,b)`` and ``(c,d)``, otherwise ``False``.

    Examples
    --------
    >>> a = Point((0,1))
    >>> b = Point((0,10))
    >>> c = Point((-2,5))
    >>> d = Point((2,5))
    >>> e = Point((-3,5))
    >>> seg_intersect(a, b, c, d)
    True

    >>> seg_intersect(a, b, c, e)
    False
    """

    segments_intersect = True

    acd_bcd = ccw([a, c, d]) == ccw([b, c, d])

    abc_abd = ccw([a, b, c]) == ccw([a, b, d])

    if acd_bcd or abc_abd:
        segments_intersect = False

    return segments_intersect


def _point_in_vertices(pt, vertices):
    """**HELPER METHOD. DO NOT CALL.** Returns whether a point
    is contained in a polygon specified by a sequence of vertices.

    Parameters
    ----------
    pt : libpysal.cg.Point
        A point.
    vertices : list
        A list of vertices representing as polygon.

    Returns
    -------
    pt_in_poly : bool
        ``True`` if ``pt`` is contained in ``vertices``, otherwise ``False``.

    Examples
    --------
    >>> _point_in_vertices(
    ...     Point((1, 1)),
    ...     [Point((0, 0)), Point((10, 0)), Point((0, 10))]
    ... )
    True
    """

    def neg_ray_intersect(p1, p2, p3) -> bool:
        """Returns whether a ray in the negative-x
        direction from ``p3`` intersects the segment between.
        """

        if not min(p1[1], p2[1]) <= p3[1] <= max(p1[1], p2[1]):
            nr_inters = False
        else:
            if p1[1] > p2[1]:
                vec1 = (p2[0] - p1[0], p2[1] - p1[1])
            else:
                vec1 = (p1[0] - p2[0], p1[1] - p2[1])

            vec2 = (p3[0] - p1[0], p3[1] - p1[1])

            nr_inters = vec1[0] * vec2[1] - vec2[0] * vec1[1] >= 0

        return nr_inters

    vert_y_set = {v[1] for v in vertices}
    while pt[1] in vert_y_set:
        # Perturb the location very slightly
        pt = pt[0], pt[1] + -1e-14 + random.random() * 2e-14

    inters = 0
    for i in range(-1, len(vertices) - 1):
        v1 = vertices[i]
        v2 = vertices[i + 1]
        if neg_ray_intersect(v1, v2, pt):
            inters += 1

    pt_in_poly = inters % 2 == 1

    return pt_in_poly


def point_touches_rectangle(point, rect):
    """Returns ``True`` (``1``) if the point is in the rectangle
    or touches it's boundary, otherwise ``False`` (``0``).

    Parameters
    ----------
    point : {libpysal.cg.Point, tuple}
        A point or point coordinates.
    rect : libpysal.cg.Rectangle
        A rectangle.

    Returns
    -------
    chflag : int
        ``1`` if ``point`` is in (or touches
        boundary of) ``rect``, otherwise ``0``.

    Examples
    --------
    >>> rect = Rectangle(0, 0, 10, 10)
    >>> a = Point((5, 5))
    >>> b = Point((10, 5))
    >>> c = Point((11, 11))
    >>> point_touches_rectangle(a, rect)
    1

    >>> point_touches_rectangle(b, rect)
    1

    >>> point_touches_rectangle(c, rect)
    0
    """

    chflag = 0
    if (point[0] >= rect.left and point[0] <= rect.right) and (
        point[1] >= rect.lower and point[1] <= rect.upper
    ):
        chflag = 1

    return chflag


def get_shared_segments(poly1, poly2, bool_ret=False):
    """Returns the line segments in common to both polygons.

    Parameters
    ----------
    poly1 : libpysal.cg.Polygon
        A Polygon.
    poly2 : libpysal.cg.Polygon
        A Polygon.
    bool_ret : bool
        Return only a ``bool``. Default is ``False``.

    Returns
    -------
    common : list
        The shared line segments between ``poly1`` and ``poly2``.
    _ret_bool : bool
        Whether ``poly1`` and ``poly2`` share a
        segment (``True``) or not (``False``).

    Examples
    --------
    >>> from libpysal.cg.shapes import Polygon
    >>> x = [0, 0, 1, 1]
    >>> y = [0, 1, 1, 0]
    >>> poly1 = Polygon(list(map(Point, zip(x, y))) )
    >>> x = [a+1 for a in x]
    >>> poly2 = Polygon(list(map(Point, zip(x, y))) )
    >>> get_shared_segments(poly1, poly2, bool_ret=True)
    True
    """

    # get_rectangle_rectangle_intersection inlined for speed.
    r0 = poly1.bounding_box
    r1 = poly2.bounding_box
    w_left = max(r0.left, r1.left)
    w_lower = max(r0.lower, r1.lower)
    w_right = min(r0.right, r1.right)
    w_upper = min(r0.upper, r1.upper)

    segments_a = set()
    common = []

    for part in poly1.parts + [p for p in poly1.holes if p]:
        if part[0] != part[-1]:  # not closed
            part = part[:] + part[0:1]
        a = part[0]

        for b in islice(part, 1, None):
            # inlining point_touches_rectangle for speed
            x, y = a
            # check if point a is in the bounding box intersection
            if x >= w_left and x <= w_right and y >= w_lower and y <= w_upper:
                x, y = b
                # check if point b is in the bounding box intersection
                if x >= w_left and x <= w_right and y >= w_lower and y <= w_upper:
                    if a > b:
                        segments_a.add((b, a))
                    else:
                        segments_a.add((a, b))
            a = b

    _ret_bool = False

    for part in poly2.parts + [p for p in poly2.holes if p]:
        if part[0] != part[-1]:  # not closed
            part = part[:] + part[0:1]
        a = part[0]

        for b in islice(part, 1, None):
            # inlining point_touches_rectangle for speed
            x, y = a
            if x >= w_left and x <= w_right and y >= w_lower and y <= w_upper:
                x, y = b
                if x >= w_left and x <= w_right and y >= w_lower and y <= w_upper:
                    seg = (b, a) if a > b else (a, b)
                    if seg in segments_a:
                        common.append(LineSegment(*seg))
                        if bool_ret:
                            _ret_bool = True
                            return _ret_bool
            a = b

    if bool_ret:
        if len(common) > 0:
            _ret_bool = True
        return _ret_bool

    return common


def distance_matrix(X, p=2.0, threshold=5e7):
    r"""Calculate a distance matrix.

    Parameters
    ----------
    X : numpy.ndarray
        An :math:`n \\times k` array where :math:`n` is the number
        of observations and :math:`k` is the number of dimensions
        (2 for :math:`x,y`).
    p : float
        Minkowski `p`-norm distance metric parameter where
        :math:`1<=\mathtt{p}<=\infty`. ``2`` is Euclidean distance and
        ``1`` is Manhattan distance. Default is ``2.0``.
    threshold : int
        If :math:`(\mathtt{n}**2)*32 > \mathtt{threshold}` use
        ``scipy.spatial.distance_matrix`` instead of working in RAM,
        this is roughly the amount of RAM (in bytes) that will be used.
        Must be positive. Default is ``5e7``.

    Returns
    -------
    d : numpy.ndarray
        An n by :math:`m` :math:`p`-norm distance matrix.

    Raises
    ------
    TypeError
        Raised when an invalid dimensional array is passed in.

    Notes
    -----
    Needs optimization/integration with other weights in PySAL.

    Examples
    --------
    >>> x, y = [r.flatten() for r in np.indices((3, 3))]
    >>> data = np.array([x, y]).T
    >>> d = distance_matrix(data)
    >>> np.array(d)
    array([[0.        , 1.        , 2.        , 1.        , 1.41421356,
            2.23606798, 2.        , 2.23606798, 2.82842712],
           [1.        , 0.        , 1.        , 1.41421356, 1.        ,
            1.41421356, 2.23606798, 2.        , 2.23606798],
           [2.        , 1.        , 0.        , 2.23606798, 1.41421356,
            1.        , 2.82842712, 2.23606798, 2.        ],
           [1.        , 1.41421356, 2.23606798, 0.        , 1.        ,
            2.        , 1.        , 1.41421356, 2.23606798],
           [1.41421356, 1.        , 1.41421356, 1.        , 0.        ,
            1.        , 1.41421356, 1.        , 1.41421356],
           [2.23606798, 1.41421356, 1.        , 2.        , 1.        ,
            0.        , 2.23606798, 1.41421356, 1.        ],
           [2.        , 2.23606798, 2.82842712, 1.        , 1.41421356,
            2.23606798, 0.        , 1.        , 2.        ],
           [2.23606798, 2.        , 2.23606798, 1.41421356, 1.        ,
            1.41421356, 1.        , 0.        , 1.        ],
           [2.82842712, 2.23606798, 2.        , 2.23606798, 1.41421356,
            1.        , 2.        , 1.        , 0.        ]])
    """

    if X.ndim == 1:
        X.shape = (X.shape[0], 1)

    if X.ndim > 2:
        msg = "Should be 2D point coordinates: %s dimensions present." % X.ndim
        raise TypeError(msg)

    n, k = X.shape

    if (n**2) * 32 > threshold:
        d = scipy.spatial.distance_matrix(X, X, p)
    else:
        m = np.ones((n, n))
        d = np.zeros((n, n))
        for col in range(k):
            x = X[:, col]
            x_m = x * m
            dx = x_m - x_m.T
            if p % 2 != 0:
                dx = np.abs(dx)
            dx2 = dx**p
            d += dx2
        d = d ** (1.0 / p)

    return d



import geopandas
import numpy as np
import pytest

from ... import examples, io, weights
from ...common import ATOL, RTOL
from .. import adjtools as adj
from ..util import lat2W


class TestAdjlist:
    def setup_method(self):
        self.knownW = io.open(examples.get_path("columbus.gal")).read()

    def test_round_trip_drop_islands_true(self):
        adjlist = self.knownW.to_adjlist(
            remove_symmetric=False, drop_islands=True
        ).astype(int)
        w_from_adj = weights.W.from_adjlist(adjlist)
        np.testing.assert_allclose(
            w_from_adj.sparse.toarray(), self.knownW.sparse.toarray()
        )

    def test_round_trip_drop_islands_false(self):
        adjlist = self.knownW.to_adjlist(
            remove_symmetric=False, drop_islands=True
        ).astype(int)
        w_from_adj = weights.W.from_adjlist(adjlist)
        np.testing.assert_allclose(
            w_from_adj.sparse.toarray(), self.knownW.sparse.toarray()
        )

    def test_filter(self):
        grid = lat2W(2, 2)
        alist = grid.to_adjlist(remove_symmetric=True, drop_islands=True)
        assert len(alist) == 4
        with pytest.raises(AssertionError):
            # build this manually because of bug libpysal#322
            alist_neighbors = alist.groupby("focal").neighbor.apply(list).to_dict()
            all_ids = set(alist_neighbors.keys()).union(
                *map(set, alist_neighbors.values())
            )
            for idx in set(all_ids).difference(set(alist_neighbors.keys())):
                alist_neighbors[idx] = []
            badgrid = weights.W(alist_neighbors)
            np.testing.assert_allclose(badgrid.sparse.toarray(), grid.sparse.toarray())
        assert set(alist.focal.unique()) == {0, 1, 2}
        assert set(alist.neighbor.unique()) == {1, 2, 3}
        assert alist.weight.unique().item() == 1
        grid = lat2W(2, 2, id_type="string")
        alist = grid.to_adjlist(remove_symmetric=True, drop_islands=True)
        assert len(alist) == 4
        with pytest.raises(AssertionError):
            # build this manually because of bug libpysal#322
            alist_neighbors = alist.groupby("focal").neighbor.apply(list).to_dict()
            all_ids = set(alist_neighbors.keys()).union(
                *map(set, alist_neighbors.values())
            )
            for idx in set(all_ids).difference(set(alist_neighbors.keys())):
                alist_neighbors[idx] = []
            badgrid = weights.W(alist_neighbors)
            np.testing.assert_allclose(badgrid.sparse.toarray(), grid.sparse.toarray())
        tuples = {tuple(t) for t in alist[["focal", "neighbor"]].values}
        full_alist = grid.to_adjlist(drop_islands=True)
        all_possible = {tuple(t) for t in full_alist[["focal", "neighbor"]].values}
        assert tuples.issubset(all_possible), (
            "the de-duped adjlist has links " "not in the duplicated adjlist."
        )
        complements = all_possible.difference(tuples)
        reversed_complements = {t[::-1] for t in complements}
        assert reversed_complements == tuples, (
            "the remaining links in the duplicated"
            " adjlist are not the reverse of the links"
            " in the deduplicated adjlist."
        )
        assert alist.weight.unique().item() == 1

    def apply_and_compare_columbus(self, col):
        import geopandas

        df = geopandas.read_file(examples.get_path("columbus.dbf")).head()
        w = weights.Queen.from_dataframe(df)
        alist = adj.adjlist_apply(df[col], W=w, to_adjlist_kws={"drop_islands": True})
        right_hovals = alist.groupby("focal").att_focal.unique()
        assert (right_hovals == df[col]).all()
        allpairs = np.subtract.outer(df[col].values, df[col].values)
        flat_diffs = allpairs[w.sparse.toarray().astype(bool)]
        np.testing.assert_allclose(flat_diffs, alist["subtract"].values)
        return flat_diffs

    def test_apply(self):
        self.apply_and_compare_columbus("HOVAL")

    def test_mvapply(self):
        import geopandas

        df = geopandas.read_file(examples.get_path("columbus.dbf")).head()
        w = weights.Queen.from_dataframe(df)

        ssq = lambda x_y: np.sum((x_y[0] - x_y[1]) ** 2).item()  # noqa: E731
        ssq.__name__ = "sum_of_squares"
        alist = adj.adjlist_apply(
            df[["HOVAL", "CRIME", "INC"]],
            W=w,
            func=ssq,
            to_adjlist_kws={"drop_islands": True},
        )
        known_ssq = [
            1301.1639302990804,
            3163.46450914361,
            1301.1639302990804,
            499.52656498472993,
            594.518273032036,
            3163.46450914361,
            499.52656498472993,
            181.79100173844196,
            436.09336916344097,
            594.518273032036,
            181.79100173844196,
            481.89443401250094,
            436.09336916344097,
            481.89443401250094,
        ]  # ugh I hate doing this, but how else?
        np.testing.assert_allclose(
            alist.sum_of_squares.values, np.asarray(known_ssq), rtol=RTOL, atol=ATOL
        )

    def test_map(self):
        atts = ["HOVAL", "CRIME", "INC"]
        df = geopandas.read_file(examples.get_path("columbus.dbf")).head()
        w = weights.Queen.from_dataframe(df)
        hoval, crime, inc = list(map(self.apply_and_compare_columbus, atts))
        mapped = adj.adjlist_map(df[atts], W=w, to_adjlist_kws={"drop_islands": True})
        for name, data in zip(atts, (hoval, crime, inc), strict=True):
            np.testing.assert_allclose(
                data, mapped["_".join(("subtract", name))].values
            )

    def test_sort(self):
        from libpysal import examples
        from libpysal.weights import Rook

        us = geopandas.read_file(examples.get_path("us48.shp"))
        w = Rook.from_dataframe(us.set_index("STATE_FIPS"), use_index=True)
        unsorted_al = w.to_adjlist(sort_joins=False)
        sorted_al = w.to_adjlist(sort_joins=True)
        sv = ["01"] * 4
        sv.append("04")
        sv = np.array(sv)
        usv = np.array(["53", "53", "30", "30", "30"])
        np.testing.assert_array_equal(unsorted_al.focal.values[:5], usv)
        np.testing.assert_array_equal(sorted_al.focal.values[:5], sv)

    def test_ids(self):
        df = geopandas.read_file(examples.get_path("columbus.dbf")).head()
        df["my_id"] = range(3, len(df) + 3)
        w = weights.Queen.from_dataframe(df, ids="my_id")
        w_adj = w.to_adjlist(drop_islands=True)
        for i in range(3, 8):
            assert i in w_adj.focal
            assert i in w_adj.neighbor
        for i in w_adj.focal:
            assert i in list(range(3, len(df) + 3))
        for i in w_adj.neighbor:
            assert i in list(range(3, len(df) + 3))

    def test_str_ids(self):
        df = geopandas.read_file(examples.get_path("columbus.dbf")).head()
        snakes = ["mamba", "boa", "python", "rattlesnake", "cobra"]
        df["my_str_id"] = snakes
        w = weights.Queen.from_dataframe(df, ids="my_str_id")
        w_adj = w.to_adjlist(drop_islands=True)
        for i in snakes:
            (w_adj.focal == i).any()
            (w_adj.neighbor == i).any()
        for i in w_adj.focal:
            assert i in snakes
        for i in w_adj.neighbor:
            assert i in snakes

    def test_lat2w(self):
        w = lat2W(5, 5)
        manual_neighbors = w.to_adjlist().groupby("focal").neighbor.agg(list).to_dict()
        for focal, neighbors in w.neighbors.items():
            assert set(manual_neighbors[focal]) == set(neighbors)



import geopandas as gpd
import numpy as np
import pandas as pd
import shapely


def _plot(
    graph_obj,
    gdf,
    focal=None,
    nodes=True,
    color="k",
    edge_kws=None,
    node_kws=None,
    focal_kws=None,
    ax=None,
    figsize=None,
    limit_extent=False,
):
    """Plot edges and nodes of the Graph

    Creates a ``maptlotlib`` plot based on the topology stored in the
    Graph and spatial location defined in ``gdf``.

    Parameters
    ----------
    graph_obj : Graph
        Graph to be plotted
    gdf : geopandas.GeoDataFrame
        Geometries indexed using the same index as Graph. Geometry types other than
        points are converted to centroids encoding start and end point of Graph
        edges.
    focal : hashable | array-like[hashable] | None, optional
        ID or an array-like of IDs of focal geometries whose weights shall be
        plotted. If None, all weights from all focal geometries are plotted.
        By default None
    nodes : bool, optional
        Plot nodes as points. Nodes are plotted using zorder=2 to show them on top of
        the edges. By default True
    color : str, optional
        The color of all objects, by default "k"
    edge_kws : dict, optional
        Keyword arguments dictionary to send to ``LineCollection``,
        which provides fine-grained control over the aesthetics
        of the edges in the plot. By default None
    node_kws : dict, optional
        Keyword arguments dictionary to send to ``ax.scatter``,
        which provides fine-grained control over the aesthetics
        of the nodes in the plot. By default None
    focal_kws : dict, optional
        Keyword arguments dictionary to send to ``ax.scatter``,
        which provides fine-grained control over the aesthetics
        of the focal nodes in the plot on top of generic ``node_kws``.
        Values of ``node_kws`` are updated from ``focal_kws``.
        Ignored if ``focal=None``. By default None
    ax : matplotlib.axes.Axes, optional
        Axis on which to plot the weights. If None, a new figure and axis are
        created. By default None
    figsize : tuple, optional
        figsize used to create a new axis. By default None
    limit_extent : bool, optional
        limit the extent of the axis to the extent of the plotted graph, by default
        False

    Returns
    -------
    matplotlib.axes.Axes
        Axis with the resulting plot
    """
    try:
        import matplotlib.pyplot as plt
        from matplotlib import collections

    except (ImportError, ModuleNotFoundError) as err:
        raise ImportError("matplotlib is required for `plot`.") from err

    if ax is None:
        f, ax = plt.subplots(figsize=figsize)

    if node_kws is not None:
        if "color" not in node_kws:
            node_kws["color"] = color
    else:
        node_kws = {"color": color}

    if edge_kws is not None:
        if "color" not in edge_kws:
            edge_kws["color"] = color
    else:
        edge_kws = {"color": color}

    # get array of coordinates in the order reflecting graph_obj._adjacency.index.codes
    # we need to work on int position to allow fast filtering of duplicated edges and
    # cannot rely on gdf remaining in the same order between Graph creation and
    # plotting
    coords = shapely.get_coordinates(gdf.reindex(graph_obj.unique_ids).centroid)

    if focal is not None:
        if not pd.api.types.is_list_like(focal):
            focal = [focal]
        subset = graph_obj._adjacency[focal]
        codes = subset.index.codes

    else:
        codes = graph_obj._adjacency.index.codes

    # avoid plotting both ij and ji
    edges = np.unique(np.sort(np.column_stack([codes]).T, axis=1), axis=0)
    lines = coords[edges]

    ax.add_collection(collections.LineCollection(lines, **edge_kws))

    if limit_extent:
        xm, ym = lines.min(axis=0).min(axis=0)
        xx, yx = lines.max(axis=0).max(axis=0)
        x_margin = (xx - xm) * 0.05
        y_margin = (yx - ym) * 0.05
        ax.set_xlim(xm - x_margin, xx + x_margin)
        ax.set_ylim(ym - y_margin, yx + y_margin)
    else:
        ax.autoscale_view()

    if nodes:
        if focal is not None:
            used_focal = coords[np.unique(subset.index.codes[0])]
            used_neighbor = coords[np.unique(subset.index.codes[1])]
            ax.scatter(used_neighbor[:, 0], used_neighbor[:, 1], **node_kws, zorder=2)
            if focal_kws is None:
                focal_kws = {}
            ax.scatter(
                used_focal[:, 0],
                used_focal[:, 1],
                **dict(node_kws, **focal_kws),
                zorder=3,
            )
        else:
            ax.scatter(coords[:, 0], coords[:, 1], **node_kws, zorder=2)

    return ax


def _explore_graph(
    g,
    gdf,
    focal=None,
    nodes=True,
    color="black",
    edge_kws=None,
    node_kws=None,
    focal_kws=None,
    m=None,
    **kwargs,
):
    """Plot graph as an interactive Folium Map

    Parameters
    ----------
    g : libpysal.Graph
        graph to be plotted
    gdf : geopandas.GeoDataFrame
        geodataframe used to instantiate to Graph
    focal : list, optional
        subset of focal observations to plot in the map, by default None.
        If none, all relationships are plotted
    nodes : bool, optional
        whether to display observations as nodes in the map, by default True
    color : str, optional
        color applied to nodes and edges, by default "black"
    edge_kws : dict, optional
        additional keyword arguments passed to geopandas explore function
        when plotting edges, by default None
    node_kws : dict, optional
        additional keyword arguments passed to geopandas explore function
        when plotting nodes, by default None
    focal_kws : dict, optional
        additional keyword arguments passed to geopandas explore function
        when plotting focal observations, by default None. Only applicable when
        passing a subset of nodes with the `focal` argument
    m : Folilum.Map, optional
        folium map objecto to plot on top of, by default None
    **kwargs : dict, optional
        additional keyword arguments are passed directly to geopandas.explore, when
        ``m=None`` by default None


    Returns
    -------
    folium.Map
        folium map
    """
    geoms = gdf.centroid.reindex(g.unique_ids)

    if node_kws is not None:
        if "color" not in node_kws:
            node_kws["color"] = color
    else:
        node_kws = {"color": color}

    if focal_kws is not None:
        if "color" not in node_kws:
            focal_kws["color"] = color
    else:
        focal_kws = {"color": color}

    if edge_kws is not None:
        if "color" not in edge_kws:
            edge_kws["color"] = color
    else:
        edge_kws = {"color": color}

    coords = shapely.get_coordinates(geoms)

    if focal is not None:
        if not pd.api.types.is_list_like(focal):
            focal = [focal]
        subset = g._adjacency[focal]
        codes = subset.index.codes
        adj = subset

    else:
        codes = g._adjacency.index.codes
        adj = g._adjacency

    # avoid plotting both ij and ji
    edges, indices = np.unique(
        np.sort(np.column_stack([codes]).T, axis=1), return_index=True, axis=0
    )
    lines = coords[edges]
    lines = gpd.GeoSeries(
        shapely.linestrings(lines),
        crs=gdf.crs,
    )
    adj = adj.iloc[indices].reset_index()
    edges = gpd.GeoDataFrame(adj, geometry=lines)[
        ["focal", "neighbor", "weight", "geometry"]
    ]

    m = (
        edges.explore(m=m, **edge_kws)
        if m is not None
        else edges.explore(**edge_kws, **kwargs)
    )

    if nodes is True:
        if focal is not None:
            # destinations
            geoms.iloc[np.unique(subset.index.codes[1])].explore(m=m, **node_kws)
            if focal_kws is None:
                focal_kws = {}
            # focals
            geoms.iloc[np.unique(subset.index.codes[0])].explore(
                m=m, **dict(node_kws, **focal_kws)
            )
        else:
            geoms.explore(m=m, **node_kws)
    return m
